import logging
from abc import abstractmethod
from threading import Lock, RLock

from .metric import Metric
from .tree_state import TreeState
from hpimdm import UnicastRouting
from .tree_interface import TreeInterface
from .tree_if_root import TreeInterfaceRoot
from .tree_if_non_root import TreeInterfaceNonRoot
from .tree_if_root_originator import TreeInterfaceRootOriginator


class KernelEntry:
    KERNEL_LOGGER = logging.getLogger('hpim.KernelEntry')

    def __init__(self, source_ip: str, group_ip: str, upstream_state_dic, interest_state_dic, kernel_entry_interface):
        self.kernel_entry_logger = logging.LoggerAdapter(self.KERNEL_LOGGER, {'tree': '(' + source_ip + ',' + group_ip + ')'})
        self.kernel_entry_logger.debug('Create KernelEntry')

        self.source_ip = source_ip
        self.group_ip = group_ip
        self._tree_state = TreeState.Inactive

        self._interest_interface_state = interest_state_dic
        self._upstream_interface_state = upstream_state_dic

        self._kernel_entry_interface = kernel_entry_interface

        ###### UNICAST INFO#################################################################
        (metric_administrative_distance, metric_cost, is_directly_connected, root_if) = \
            UnicastRouting.get_unicast_info(source_ip)
        # TODO verificar is directly connected
        self._rpc = Metric(metric_administrative_distance, metric_cost)
        #######################################################################################
        # Locks
        self._multicast_change = Lock()
        self._lock_test2 = RLock()
        self.CHANGE_STATE_LOCK = RLock()

        # select root interface based on rpf check
        self.inbound_interface_index = root_if
        self.interface_state = {}  # type: dict(int, TreeInterface)

    def get_inbound_interface_index(self):
        """
        Get VIF of root interface of this tree
        """
        return self.inbound_interface_index

    def get_outbound_interfaces_indexes(self):
        """
        Get OIL of this tree
        """
        return self._kernel_entry_interface.get_outbound_interfaces_indexes(self)

    @abstractmethod
    def check_tree_state(self):
        """
        Verify if tree changes state (Active/Unsure/Inactive)
        """
        return

    ################################################
    # Receive (S,G) data packets or control packets
    ################################################
    def recv_data_msg(self, index):
        """
        Receive data packet regarding this tree in interface with VIF index
        """
        print("recv data")
        self.interface_state[index].recv_data_msg()

    ###############################################################
    # Code related with tree state
    ###############################################################
    def check_interface_state(self, index, upstream_state, interest_state):
        """
        A neighbor changed Upstream state due to the reception of any control packet
        (IamUpstream or IamNoLongerUpstream or Interest or NoInterest or Sync)
        """
        if index not in self.interface_state or self.is_tree_inactive():
            return
        print("ENTER CHECK INTERFACE STATE")
        self._upstream_interface_state[index] = upstream_state

        self.interface_state[index].change_best_upstream_neighbor_state(upstream_state)
        self.check_interest_state(index, interest_state)

        self.check_tree_state()
        print("EXIT CHECK INTERFACE STATE")

    def check_interest_state(self, index, interest_state):
        """
        A neighbor changed Interest state due to the reception of any control packet
        (IamUpstream or IamNoLongerUpstream or Interest or NoInterest or Sync)
        """
        if index not in self.interface_state or self.is_tree_inactive():
            return

        current_interest_state = self._interest_interface_state.get(index, None)
        self._interest_interface_state[index] = interest_state

        if current_interest_state != interest_state:
            self.interface_state[index].change_interest_state(interest_state)

    def check_membership_state(self, index):
        """
        Reverify IGMP/MLD state of this tree in interface with VIF index...
        This is invoked whenever interface index enables or disables IGMP/MLD
        """
        print("ENTER CHECK IGMP STATE")
        if index not in self.interface_state:
            return

        self.interface_state[index].check_membership_state()
        print("EXIT CHECK IGMP STATE")

    def get_interface_sync_state(self, vif_index):
        """
        Determine if this tree must be included in a new snapshot of interface with VIF vif_index
        """
        with self.CHANGE_STATE_LOCK:
            if vif_index not in self.interface_state:
                return None
            else:
                return self.interface_state[vif_index].get_sync_state()

    def is_tree_active(self):
        """
        Determine if tree is in Active state
        """
        return self._tree_state.is_active()

    def is_tree_unsure(self):
        """
        Determine if tree is in Unsure state
        """
        return self._tree_state.is_unsure()

    def is_tree_inactive(self):
        """
        Determine if tree is in Inactive state
        """
        return self._tree_state.is_inactive()

    def set_tree_state(self, tree_state):
        """
        Set tree state (Active/Unsure/Inactive)
        """
        with self.CHANGE_STATE_LOCK:
            self.kernel_entry_logger.debug('Tree transitions to ' + str(tree_state))
            self._tree_state = tree_state

    ###############################################################
    # Unicast Changes to RPF
    ###############################################################
    @abstractmethod
    def network_update(self):
        """
        Router suffered an RPC chage... React to this change
        """
        return

    def is_in_tree(self):
        """
        Determine if router is interested in receiving data packets
        """
        for interface in self.interface_state.values():
            if interface.is_forwarding():
                return True
        return False

    @abstractmethod
    def evaluate_in_tree_change(self):
        """
        Evaluate if there is a change of interest from this router
        """
        return

    def get_source(self):
        """
        Get source IP of this tree
        """
        return self.source_ip

    def get_group(self):
        """
        Get group IP of this tree
        """
        return self.group_ip

    def change(self):
        """
        Reset multicast routing table due to changes in state
        """
        with self._multicast_change:
            if self.inbound_interface_index is not None and not self.is_tree_inactive():
                self.get_kernel().set_multicast_route(self)

    def remove_entry(self):
        """
        Remove entry from the multicast routing table
        """
        self.get_kernel().remove_multicast_route(self)

    def delete_state(self):
        """
        Delete all stored state regarding this tree
        """
        for state in self.interface_state.values():
            state.delete()
        self.interface_state.clear()

    def get_interface_name(self, interface_id):
        """
        Get name of interface from vif id
        """
        return self._kernel_entry_interface.get_interface_name(interface_id)

    def get_interface(self, interface_id):
        """
        Get HPIM interface from interface id
        """
        return self._kernel_entry_interface.get_interface(self, interface_id)

    def get_membership_interface(self, interface_id):
        """
        Get IGMP/MLD interface from interface id
        """
        return self._kernel_entry_interface.get_membership_interface(self, interface_id)

    def get_kernel(self):
        """
        Get kernel
        """
        return self._kernel_entry_interface.get_kernel()

    ######################################
    # Interface change
    #######################################
    @abstractmethod
    def new_interface(self, index):
        """
        New interface with VIF index added
        """
        return

    def remove_interface(self, index):
        """
        Interface with VIF index removed
        """
        with self.CHANGE_STATE_LOCK:
            if index not in self.interface_state:
                return

            #check if removed interface is root interface
            if self.inbound_interface_index == index:
                self.inbound_interface_index = None

            # remove cached info about removed interface
            self._upstream_interface_state.pop(index, None)
            self._interest_interface_state.pop(index, None)

            self.interface_state.pop(index).delete()
            self.change()
            self.evaluate_in_tree_change()
            self.check_tree_state()


class KernelEntryNonOriginator(KernelEntry):
    def __init__(self, source_ip: str, group_ip: str, upstream_state_dic, interest_state_dic, kernel_entry_interface):
        super().__init__(source_ip, group_ip, upstream_state_dic, interest_state_dic, kernel_entry_interface)

        # (S,G) starts OUT-TREE state... later check if node is in-tree via evaluate_in_tree_change()
        self._was_in_tree = False

        self.first_check_tree_state()
        with self.CHANGE_STATE_LOCK:
            for i in self.get_kernel().vif_index_to_name_dic.keys():
                try:
                    upstream_state = self._upstream_interface_state.get(i, None)

                    if i == self.inbound_interface_index:
                        continue
                    else:
                        interest_state = self._interest_interface_state.get(i, False)
                        self.interface_state[i] = TreeInterfaceNonRoot(self, i, self._rpc,
                                                                       best_upstream_router=upstream_state,
                                                                       interest_state=interest_state,
                                                                       was_root=False,
                                                                       previous_tree_state=TreeState.Inactive,
                                                                       current_tree_state=self._tree_state)

                except:
                    import traceback
                    print(traceback.print_exc())
                    continue
            self._was_in_tree = self.is_in_tree()
            upstream_state = self._upstream_interface_state.get(self.inbound_interface_index, None)
            if self.inbound_interface_index is not None:
                self.interface_state[self.inbound_interface_index] = \
                    TreeInterfaceRoot(self, self.inbound_interface_index, upstream_state, was_non_root=False,
                                      previous_tree_state=TreeState.Inactive, current_tree_state=self._tree_state)

        self.change()
        self.evaluate_in_tree_change()
        print('Tree NonOriginator created')

    def check_tree_state(self):
        """
        Verify if tree changes state (Active/Unsure/Inactive)
        """
        with self.CHANGE_STATE_LOCK:
            if self.inbound_interface_index is not None and \
                    len(self.interface_state) > 0 and\
                    self._upstream_interface_state.get(self.inbound_interface_index, None) is not None and \
                    (not self._rpc.is_better_than(self._upstream_interface_state.get(self.inbound_interface_index))
                     and self._rpc != self._upstream_interface_state.get(self.inbound_interface_index)):
                # tree is Active
                print("PARA ACTIVE")
                self._tree_state.transition_to_active(self)
            elif len(self.interface_state) > 0 and \
                    not all(value is None for value in self._upstream_interface_state.values()):
                print("PARA UNSURE")
                self._tree_state.transition_to_unsure(self)
            else:
                print("PARA INACTIVE")
                self._tree_state.transition_to_inactive(self)

    def first_check_tree_state(self):
        """
        Verify for the first time in which state the tree is at
        """
        if self.inbound_interface_index is not None and \
                self._upstream_interface_state.get(self.inbound_interface_index, None) is not None and \
                (not self._rpc.is_better_than(self._upstream_interface_state.get(self.inbound_interface_index))
                 and self._rpc != self._upstream_interface_state.get(self.inbound_interface_index)):
            # tree is Active
            print("PARA ACTIVE")
            self.set_tree_state(TreeState.Active)
        elif not all(value is None for value in self._upstream_interface_state.values()):
            print("PARA UNSURE")
            # tree is Unsure
            self.set_tree_state(TreeState.Unsure)
        else:
            print("PARA INACTIVE")
            # tree is Inactive
            self.set_tree_state(TreeState.Inactive)

    ###############################################################
    # Unicast Changes to RPF
    ###############################################################
    def network_update(self):
        """
        Router suffered an RPC chage... React to this change
        """
        with self.CHANGE_STATE_LOCK:
            (metric_administrative_distance, metric_cost, is_directly_connected, new_inbound_interface_index) = \
                UnicastRouting.get_unicast_info(self.source_ip)
            new_rpc = Metric(metric_administrative_distance, metric_cost)

            if is_directly_connected:
                self._tree_state.transition_to_inactive(self)
                return
            if new_inbound_interface_index != self.inbound_interface_index:
                # get old interfaces
                old_upstream_interface = self.interface_state.get(self.inbound_interface_index, None)
                old_downstream_interface = self.interface_state.get(new_inbound_interface_index, None)

                non_root_interest_state = self._interest_interface_state.get(self.inbound_interface_index, False)
                non_root_upstream_state = self._upstream_interface_state.get(self.inbound_interface_index, None)
                root_interest_state = self._interest_interface_state.get(new_inbound_interface_index, False)
                root_upstream_state = self._upstream_interface_state.get(new_inbound_interface_index, None)

                # remove old interfaces
                if old_upstream_interface is not None:
                    old_upstream_interface.delete()
                if old_downstream_interface is not None:
                    old_downstream_interface.delete()

                new_tree_state = self._tree_state
                if (self.is_tree_active() and root_upstream_state is None) or \
                        (self.is_tree_active() and root_upstream_state is not None and
                         (new_rpc.is_better_than(root_upstream_state) or new_rpc == root_upstream_state)):
                    new_tree_state = TreeState.Unsure
                elif self.is_tree_unsure() and root_upstream_state is not None and \
                        not new_rpc.is_better_than(root_upstream_state) and new_rpc != root_upstream_state:
                    new_tree_state = TreeState.Active

                # change type of interfaces
                if self.inbound_interface_index is not None:
                    new_downstream_interface = TreeInterfaceNonRoot(self, self.inbound_interface_index, new_rpc,
                                                                    non_root_upstream_state, non_root_interest_state,
                                                                    was_root=True,
                                                                    previous_tree_state=self._tree_state,
                                                                    current_tree_state=new_tree_state)
                    self.interface_state[self.inbound_interface_index] = new_downstream_interface
                if new_inbound_interface_index is not None:
                    new_upstream_interface = TreeInterfaceRoot(self, new_inbound_interface_index,
                                                               root_upstream_state, was_non_root=True,
                                                               previous_tree_state=self._tree_state,
                                                               current_tree_state=new_tree_state)
                    self.interface_state[new_inbound_interface_index] = new_upstream_interface
                self.inbound_interface_index = new_inbound_interface_index

                if self._rpc != new_rpc:
                    self._rpc = new_rpc
                    self.check_tree_state()
                    for interface in self.interface_state.values():
                        interface.notify_rpc_change(new_rpc)
                else:
                    self.check_tree_state()

                # atualizar tabela de encaminhamento multicast
                self.change()
                self.evaluate_in_tree_change()
            elif self._rpc != new_rpc:
                self._rpc = new_rpc
                self.check_tree_state()
                for interface in self.interface_state.values():
                    interface.notify_rpc_change(new_rpc)

    def evaluate_in_tree_change(self):
        """
        Evaluate if there is a change of interest from this router
        """
        with self._lock_test2:
            is_in_tree = self.is_in_tree()
            was_in_tree = self._was_in_tree
            self._was_in_tree = is_in_tree
            if was_in_tree != is_in_tree and self.inbound_interface_index is not None:
                if is_in_tree:
                    self.interface_state[self.inbound_interface_index].node_is_in_tree()
                else:
                    self.interface_state[self.inbound_interface_index].node_is_out_tree()

    #####################################################
    # New interface configured
    #####################################################
    def new_interface(self, index):
        """
        New interface with VIF index added
        """
        print("NEW_INTERFACE ANTES")
        with self.CHANGE_STATE_LOCK:
            print("NEW_INTERFACE DEPOIS")
            if index in self.interface_state:
                return

            (_, _, _, inbound_interface_index) = UnicastRouting.get_unicast_info(self.source_ip)
            # TODO verificar is directly connected

            interest_state = False
            upstream_state = None
            self._interest_interface_state[index] = interest_state
            self._upstream_interface_state[index] = upstream_state

            # new interface is of type non-root
            if inbound_interface_index != index:
                self.interface_state[index] = TreeInterfaceNonRoot(self, index, self._rpc,
                                                                   best_upstream_router=upstream_state,
                                                                   interest_state=interest_state,
                                                                   was_root=False,
                                                                   previous_tree_state=self._tree_state,
                                                                   current_tree_state=self._tree_state)
            # new interface is of type root and there wasnt any root interface previously configured
            elif inbound_interface_index == index and self.inbound_interface_index is None:
                self.inbound_interface_index = index
                self.interface_state[index] = TreeInterfaceRoot(self, self.inbound_interface_index,
                                                                upstream_state, was_non_root=False,
                                                                previous_tree_state=self._tree_state,
                                                                current_tree_state=self._tree_state)
            # new interface is of type root and there was a root interface previously configured
            elif inbound_interface_index == index and self.inbound_interface_index is not None:
                old_upstream_interface = self.interface_state.get(self.inbound_interface_index, None)

                root_upstream_state = upstream_state
                non_root_interest_state = self._interest_interface_state.get(self.inbound_interface_index, False)
                non_root_upstream_state = self._upstream_interface_state.get(self.inbound_interface_index, None)

                new_tree_state = self._tree_state
                if self.is_tree_active() and root_upstream_state is None:
                    new_tree_state = TreeState.Unsure

                # change type of interfaces
                if self.inbound_interface_index is not None:
                    new_downstream_interface = TreeInterfaceNonRoot(self, self.inbound_interface_index, self._rpc,
                                                                    non_root_upstream_state, non_root_interest_state,
                                                                    was_root=True,
                                                                    previous_tree_state=self._tree_state,
                                                                    current_tree_state=new_tree_state)
                    self.interface_state[self.inbound_interface_index] = new_downstream_interface
                if inbound_interface_index is not None:
                    new_upstream_interface = TreeInterfaceRoot(self, inbound_interface_index, root_upstream_state,
                                                               was_non_root=True,
                                                               previous_tree_state=self._tree_state,
                                                               current_tree_state=new_tree_state)
                    self.interface_state[inbound_interface_index] = new_upstream_interface
                self.inbound_interface_index = inbound_interface_index

                # remove old interfaces
                if old_upstream_interface is not None:
                    old_upstream_interface.delete()

            self.change()
            self.evaluate_in_tree_change()
            self.check_tree_state()


class KernelEntryOriginator(KernelEntry):
    KERNEL_LOGGER = logging.getLogger('hpim.KernelEntryOriginator')

    def __init__(self, source_ip: str, group_ip: str, upstream_state_dic, interest_state_dic, kernel_entry_interface):
        super().__init__(source_ip, group_ip, upstream_state_dic, interest_state_dic, kernel_entry_interface)
        self.sat_is_running = False
        if self.inbound_interface_index is not None:
            self.sat_is_running = True
            self.set_tree_state(TreeState.Active)
        else:
            self.set_tree_state(TreeState.Unsure)

        with self.CHANGE_STATE_LOCK:
            for i in self.get_kernel().vif_index_to_name_dic.keys():
                try:
                    upstream_state = self._upstream_interface_state.get(i, None)

                    if i != self.inbound_interface_index:
                        interest_state = self._interest_interface_state.get(i, False)
                        self.interface_state[i] = TreeInterfaceNonRoot(self, i, self._rpc,
                                                                       best_upstream_router=upstream_state,
                                                                       interest_state=interest_state,
                                                                       was_root=False,
                                                                       previous_tree_state=TreeState.Inactive,
                                                                       current_tree_state=self._tree_state)
                except:
                    import traceback
                    print(traceback.print_exc())
                    continue

            if self.inbound_interface_index is not None:
                self.interface_state[self.inbound_interface_index] = \
                    TreeInterfaceRootOriginator(self, self.inbound_interface_index, self._tree_state)

        self.change()
        self.check_tree_state()
        print('Tree Originator created')

    def check_tree_state(self):
        """
        Verify if tree changes state (Active/Unsure/Inactive)
        """
        if self.inbound_interface_index is not None and self.sat_is_running and len(self.interface_state) > 0:
            print("PARA ACTIVE")
            # tree is active
            self._tree_state.transition_to_active(self)
        elif len(self.interface_state) == 0 or\
                (not self.sat_is_running and all(v is None for v in self._upstream_interface_state.values())):
            # tree is Inactive
            print("PARA INACTIVE")
            self._tree_state.transition_to_inactive(self)
        elif self.inbound_interface_index is None or not self.sat_is_running:
            print("PARA UNSURE")
            # tree is Unsure
            self._tree_state.transition_to_unsure(self)

    ###############################################################
    # Code related with tree state
    ###############################################################
    def sat_expires(self):
        """
        Source Active has expired
        """
        self.sat_is_running = False
        self.check_tree_state()

    def sat_running(self):
        """
        Source Active timer was set
        """
        self.sat_is_running = True
        self.check_tree_state()

    ###############################################################
    # Unicast Changes to RPF
    ###############################################################
    def network_update(self):
        """
        Router suffered an RPC chage... React to this change
        """
        with self.CHANGE_STATE_LOCK:
            (metric_administrative_distance, metric_cost, is_directly_connected, new_inbound_interface_index) = \
                UnicastRouting.get_unicast_info(self.source_ip)
            new_rpc = Metric(metric_administrative_distance, metric_cost)

            if not is_directly_connected:
                self._tree_state.transition_to_inactive(self)
                return
            if new_inbound_interface_index != self.inbound_interface_index:
                # get old interfaces
                old_upstream_interface = self.interface_state.get(self.inbound_interface_index, None)
                old_downstream_interface = self.interface_state.get(new_inbound_interface_index, None)

                non_root_interest_state = self._interest_interface_state.get(self.inbound_interface_index, False)
                non_root_upstream_state = self._upstream_interface_state.get(self.inbound_interface_index, None)
                root_interest_state = self._interest_interface_state.get(new_inbound_interface_index, False)
                root_upstream_state = self._upstream_interface_state.get(new_inbound_interface_index, None)

                if new_inbound_interface_index is None:
                    self.sat_is_running = False
                    new_tree_state = TreeState.Unsure
                else:
                    new_tree_state = TreeState.Active

                # change type of interfaces
                if self.inbound_interface_index is not None:
                    new_downstream_interface = TreeInterfaceNonRoot(self, self.inbound_interface_index, self._rpc,
                                                                    non_root_upstream_state, non_root_interest_state,
                                                                    was_root=True,
                                                                    previous_tree_state=self._tree_state,
                                                                    current_tree_state=new_tree_state)
                    self.interface_state[self.inbound_interface_index] = new_downstream_interface
                if new_inbound_interface_index is not None:
                    self.sat_is_running = True
                    new_upstream_interface = TreeInterfaceRootOriginator(self, new_inbound_interface_index, new_tree_state)
                    self.interface_state[new_inbound_interface_index] = new_upstream_interface
                self.inbound_interface_index = new_inbound_interface_index

                # remove old interfaces
                if old_upstream_interface is not None:
                    old_upstream_interface.delete()
                if old_downstream_interface is not None:
                    old_downstream_interface.delete()

                self.check_tree_state()
                self.change()
            if self._rpc != new_rpc:
                self._rpc = new_rpc
                for interface in self.interface_state.values():
                    interface.notify_rpc_change(new_rpc)

    def evaluate_in_tree_change(self):
        """
        Evaluate if there is a change of interest from this router
        """
        return

    #####################################################
    # New interface configured
    #####################################################
    def new_interface(self, index):
        """
        New interface with VIF index added
        """
        print("NEW_INTERFACE ANTES")
        with self.CHANGE_STATE_LOCK:
            print("NEW_INTERFACE DEPOIS")
            if index in self.interface_state:
                return

            (_, _, _, inbound_interface_index) = UnicastRouting.get_unicast_info(self.source_ip)
            # TODO verificar is directly connected
            interest_state = False
            upstream_state = None
            self._interest_interface_state[index] = interest_state
            self._upstream_interface_state[index] = upstream_state

            if inbound_interface_index != index:
                self.interface_state[index] = TreeInterfaceNonRoot(self, index, self._rpc,
                                                                   best_upstream_router=upstream_state,
                                                                   interest_state=interest_state,
                                                                   was_root=False,
                                                                   previous_tree_state=self._tree_state,
                                                                   current_tree_state=self._tree_state)
            elif inbound_interface_index == index and self.inbound_interface_index is None:
                self.inbound_interface_index = index
                self.sat_is_running = True
                self.interface_state[index] = \
                    TreeInterfaceRootOriginator(self, self.inbound_interface_index, self._tree_state)
            elif inbound_interface_index == index and self.inbound_interface_index is not None:
                old_upstream_interface = self.interface_state.get(self.inbound_interface_index, None)
                non_root_interest_state = self._interest_interface_state.get(self.inbound_interface_index, False)
                non_root_upstream_state = self._upstream_interface_state.get(self.inbound_interface_index, None)

                # change type of interfaces
                if self.inbound_interface_index is not None:
                    new_downstream_interface = TreeInterfaceNonRoot(self, self.inbound_interface_index, self._rpc,
                                                                    non_root_upstream_state, non_root_interest_state,
                                                                    was_root=True,
                                                                    previous_tree_state=self._tree_state,
                                                                    current_tree_state=self._tree_state)
                    self.interface_state[self.inbound_interface_index] = new_downstream_interface
                if inbound_interface_index is not None:
                    self.sat_is_running = True
                    new_upstream_interface = TreeInterfaceRootOriginator(self, inbound_interface_index,
                                                                         self._tree_state)
                    self.interface_state[inbound_interface_index] = new_upstream_interface
                self.inbound_interface_index = inbound_interface_index

                # remove old interfaces
                if old_upstream_interface is not None:
                    old_upstream_interface.delete()

            self.change()
            self.evaluate_in_tree_change()

    def remove_interface(self, index):
        """
        Interface with VIF index removed
        """
        with self.CHANGE_STATE_LOCK:
            super().remove_interface(index)
            if self.inbound_interface_index is None:
                self.sat_is_running = False
                self.check_tree_state()



from abc import ABCMeta, abstractmethod


class LocalMembershipStateABC(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def has_members():
        """
        Determine if this state considers to have hosts interested in receiving data packets
        """
        raise NotImplementedError


class NoInfo(LocalMembershipStateABC):
    @staticmethod
    def has_members():
        """
        Determine if this state considers to have hosts interested in receiving data packets
        """
        return False

    def __str__(self):
        return "NoInfo"


class Include(LocalMembershipStateABC):
    @staticmethod
    def has_members():
        """
        Determine if this state considers to have hosts interested in receiving data packets
        """
        return True

    def __str__(self):
        return "Include"


class LocalMembership():
    NoInfo = NoInfo()
    Include = Include()



import time
import socket
import struct
import logging
import ipaddress
import netifaces
from threading import RLock
from socket import if_nametoindex

from hpimdm import Main
from .packet.Packet import Packet
from .Interface import Interface
from .InterfaceHPIM import InterfaceHPIM
from .packet.ReceivedPacket import ReceivedPacket_v6
from hpimdm.tree.hpim_globals import MSG_FORMAT
if MSG_FORMAT == "BINARY":
    from hpimdm.packet.PacketHPIMHeader import PacketHPIMHeader_v6
    from hpimdm.packet.PacketHPIMIamUpstream import PacketHPIMUpstream_v6
    from hpimdm.packet.PacketHPIMNotUpstream import PacketHPIMNoLongerUpstream_v6
    from hpimdm.packet.PacketHPIMInterest import PacketHPIMInterest_v6
    from hpimdm.packet.PacketHPIMInterest import PacketHPIMNoInterest_v6
    from hpimdm.packet.PacketHPIMSync import PacketHPIMSyncEntry_v6, PacketHPIMSync_v6
    from hpimdm.packet.PacketHPIMAck import PacketHPIMAck_v6
else:
    from hpimdm.packet.PacketHPIMHeader import PacketHPIMHeaderJson as PacketHPIMHeader_v6
    from hpimdm.packet.PacketHPIMIamUpstream import PacketHPIMUpstreamJson as PacketHPIMUpstream_v6
    from hpimdm.packet.PacketHPIMNotUpstream import PacketHPIMNoLongerUpstreamJson as PacketHPIMNoLongerUpstream_v6
    from hpimdm.packet.PacketHPIMInterest import PacketHPIMInterestJson as PacketHPIMInterest_v6
    from hpimdm.packet.PacketHPIMInterest import PacketHPIMNoInterestJson as PacketHPIMNoInterest_v6
    from hpimdm.packet.PacketHPIMSync import PacketHPIMSyncEntryJson as PacketHPIMSyncEntry_v6,\
        PacketHPIMSyncJson as PacketHPIMSync_v6
    from hpimdm.packet.PacketHPIMAck import PacketHPIMAckJson as PacketHPIMAck_v6


class InterfaceHPIM6(InterfaceHPIM):
    MCAST_GRP = "ff02::d"

    def __init__(self, interface_name: str, vif_index: int):
        self.interface_logger = logging.LoggerAdapter(InterfaceHPIM.LOGGER, {'vif': vif_index,
                                                                             'interfacename': interface_name})

        # Generate BootTime
        self.time_of_boot = int(time.time())

        # Regulate transmission of Hello messages
        self.hello_timer = None

        # protocol neighbors
        self._had_neighbors = False
        self.neighbors = {}
        self.neighbors_lock = RLock()

        # reliable transmission buffer
        self.reliable_transmission_buffer = {} # Key: ID da msg ; value: ReliableMsgTransmission
        self.reliable_transmission_lock = RLock()

        # sequencer for msg reliability
        self.sequencer = 0
        self.sequencer_lock = RLock()

        # security
        self.security_id = 0
        self.security_len = 0
        self.hash_function = None
        self.security_key = b''

        # SOCKET
        s = socket.socket(socket.AF_INET6, socket.SOCK_RAW, socket.IPPROTO_PIM)

        ip_interface = "::"
        for if_addr in netifaces.ifaddresses(interface_name)[netifaces.AF_INET6]:
            ip_interface = if_addr["addr"]
            if ipaddress.IPv6Address(ip_interface.split("%")[0]).is_link_local:
                # bind to interface
                s.bind(socket.getaddrinfo(ip_interface, None, 0, socket.SOCK_RAW, 0, socket.AI_PASSIVE)[0][4])
                ip_interface = ip_interface.split("%")[0]
                break

        self.ip_interface = ip_interface

        # allow other sockets to bind this port too
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # explicitly join the multicast group on the interface specified
        if_index = if_nametoindex(interface_name)
        s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP,
                     socket.inet_pton(socket.AF_INET6, InterfaceHPIM6.MCAST_GRP) + struct.pack('@I', if_index))
        s.setsockopt(socket.SOL_SOCKET, 25, str(interface_name + '\0').encode('utf-8'))

        # set socket output interface
        s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_IF, struct.pack('@I', if_index))

        # set socket TTL to 1
        s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_HOPS, 1)
        s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_UNICAST_HOPS, 1)

        # don't receive outgoing packets
        s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_LOOP, 0)
        Interface.__init__(self, interface_name, s, s, vif_index)
        self.force_send_hello()

    @staticmethod
    def get_kernel():
        """
        Get Kernel object
        """
        return Main.kernel_v6

    @staticmethod
    def _get_address_family():
        """
        Get address family for this interface
        """
        return socket.AF_INET6

    @staticmethod
    def get_ip_header_length():
        """"
        Method for getting fixed IPv6 header length. Useful for doing calculations with MTU
        """
        return 40

    def send(self, data: bytes, group_ip: str=MCAST_GRP):
        """
        Send a new packet destined to group_ip IP
        """
        super().send(data=data, group_ip=group_ip)

    def _receive(self, raw_bytes, ancdata, src_addr):
        """
        Interface received a new control packet
        """
        if raw_bytes:
            packet = ReceivedPacket_v6(raw_bytes, ancdata, src_addr, 103, self)
            self.PKT_FUNCTIONS[packet.payload.get_pim_type()](self, packet)

    ###########################################
    # Create packets
    ###########################################
    def create_i_am_upstream_msg(self, my_boot_time, sn, source, group, metric_preference, metric):
        ph = PacketHPIMUpstream_v6(source, group, metric_preference, metric, sn)
        return Packet(payload=PacketHPIMHeader_v6(payload=ph, boot_time=my_boot_time))

    def create_i_am_no_longer_upstream_msg(self, my_boot_time, sn, source, group):
        ph = PacketHPIMNoLongerUpstream_v6(source, group, sn)
        return Packet(payload=PacketHPIMHeader_v6(payload=ph, boot_time=my_boot_time))

    def create_interest_msg(self, my_boot_time, sn, source, group):
        ph = PacketHPIMInterest_v6(source, group, sn)
        return Packet(payload=PacketHPIMHeader_v6(payload=ph, boot_time=my_boot_time))

    def create_no_interest_msg(self, my_boot_time, sn, source, group):
        ph = PacketHPIMNoInterest_v6(source, group, sn)
        return Packet(payload=PacketHPIMHeader_v6(payload=ph, boot_time=my_boot_time))

    def create_ack_msg(self, my_boot_time, sn, source, group, neighbor_boot_time,
                       neighbor_snapshot_sn, my_snapshot_sn):
        ack = PacketHPIMAck_v6(source_ip=source, group_ip=group, sequence_number=sn,
                               neighbor_boot_time=neighbor_boot_time, neighbor_snapshot_sn=neighbor_snapshot_sn,
                               my_snapshot_sn=my_snapshot_sn)
        return Packet(payload=PacketHPIMHeader_v6(payload=ack, boot_time=my_boot_time))

    def create_sync_entry_hdr(self, source, group, metric_preference, metric):
        return PacketHPIMSyncEntry_v6(source, group, metric_preference, metric)

    def create_sync_msg(self, my_boot_time, my_snapshot_sn, neighbor_snapshot_sn, sync_sn, upstream_trees,
                        master_flag, more_flag, neighbor_boot_time):
        pkt_sync = PacketHPIMSync_v6(my_snapshot_sn, neighbor_snapshot_sn, sync_sn, upstream_trees,
                                     master_flag, more_flag, neighbor_boot_time)
        return Packet(payload=PacketHPIMHeader_v6(pkt_sync, my_boot_time))



import struct
import socket
import ipaddress

###########################################################################################################
# JSON FORMAT
###########################################################################################################
class PacketHPIMUpstreamJson():
    PIM_TYPE = "I_AM_UPSTREAM"

    def __init__(self, source, group, metric_preference, metric, sequence_number):
        self.source = source
        self.group = group
        self.metric = metric
        self.metric_preference = metric_preference
        self.sequence_number = sequence_number

    def bytes(self) -> bytes:
        """
        Obtain Protocol IamUpstream Packet in a format to be transmitted (JSON)
        """
        msg = {"SOURCE": self.source,
               "GROUP": self.group,
               "METRIC": self.metric,
               "METRIC_PREFERENCE": self.metric_preference,
               "SN": self.sequence_number
              }

        return msg

    @classmethod
    def parse_bytes(cls, data: bytes):
        """
        Parse received Protocol IamUpstream Packet from JSON format and convert it into ProtocolUpstream object
        """
        source = data["SOURCE"]
        group = data["GROUP"]
        metric = data["METRIC"]
        metric_preference = data["METRIC_PREFERENCE"]
        sn = data["SN"]
        return cls(source, group, metric_preference, metric, sn)


###########################################################################################################
# BINARY FORMAT
###########################################################################################################
'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                        Tree Source IP                         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Tree Group IP                         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                        Sequence Number                        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                       Metric Preference                       |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                            Metric                             |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''
class PacketHPIMUpstream:
    PIM_TYPE = 2

    PIM_HDR_INSTALL = "! 4s 4s L L L"
    PIM_HDR_INSTALL_LEN = struct.calcsize(PIM_HDR_INSTALL)
    FAMILY = socket.AF_INET

    def __init__(self, source_ip, group_ip, metric_preference, metric, sequence_number):
        if type(source_ip) not in (str, bytes) or type(group_ip) not in (str, bytes):
            raise Exception
        if type(source_ip) is bytes:
            source_ip = socket.inet_ntop(self.FAMILY, source_ip)
        if type(group_ip) is bytes:
            group_ip = socket.inet_ntop(self.FAMILY, group_ip)

        self.source = source_ip
        self.group = group_ip
        self.metric = metric
        self.metric_preference = metric_preference
        self.sequence_number = sequence_number

    def bytes(self) -> bytes:
        """
        Obtain Protocol IamUpstream Packet in a format to be transmitted (binary)
        """
        msg = struct.pack(self.PIM_HDR_INSTALL, socket.inet_pton(self.FAMILY, self.source),
                          socket.inet_pton(self.FAMILY, self.group), self.sequence_number,
                          self.metric_preference, self.metric)

        return msg

    def __len__(self):
        return len(self.bytes())

    @classmethod
    def parse_bytes(cls, data: bytes):
        """
        Parse received Protocol IamUpstream Packet from binary format and convert it into ProtocolUpstream object
        """
        (tree_source, tree_group, sn, mp, m) = struct.unpack(cls.PIM_HDR_INSTALL,
                                                             data[:cls.PIM_HDR_INSTALL_LEN])
        return cls(tree_source, tree_group, mp, m, sn)


class PacketHPIMUpstream_v6(PacketHPIMUpstream):
    PIM_HDR_INSTALL = "! 16s 16s L L L"
    PIM_HDR_INSTALL_LEN = struct.calcsize(PIM_HDR_INSTALL)
    FAMILY = socket.AF_INET6

    def __init__(self, source_ip, group_ip, metric_preference, metric, sequence_number):
        super().__init__(source_ip, group_ip, metric_preference, metric, sequence_number)



import logging
import ipaddress
from threading import RLock
from abc import ABCMeta, abstractmethod

from .tree_state import TreeState
from .local_membership import LocalMembership


class TreeInterface(metaclass=ABCMeta):
    def __init__(self, kernel_entry, interface_id, best_upstream_router, current_tree_state, logger: logging.LoggerAdapter):
        self._kernel_entry = kernel_entry
        self._interface_id = interface_id
        self.logger = logger

        self._best_upstream_router = best_upstream_router  # current assert winner

        self.current_tree_state = current_tree_state  # current tree state

        # Local Membership State
        self._igmp_lock = RLock()
        try:
            membership_interface = self.get_membership_interface()  # type: InterfaceIGMP
            group_state = membership_interface.interface_state.get_group_state(kernel_entry.group_ip)
            has_members = group_state.add_multicast_routing_entry(self)
            self._local_membership_state = LocalMembership.Include if has_members else LocalMembership.NoInfo
        except:
            self._local_membership_state = LocalMembership.NoInfo

    ###########################################
    # Recv packets
    ###########################################
    def recv_data_msg(self):
        """
        This interface received a data packet
        """
        pass

    ######################################
    # Send messages
    ######################################
    def get_sync_state(self):
        """
        Determine if this tree must be included in a new snapshot
        By default not include this tree in snapshot... This behavior is overrode by subclasses (in NonRoot interfaces)
        """
        return None

    def send_i_am_no_longer_upstream(self):
        """
        Send an IamNoLongerUpstream message through this interface
        """
        (source, group) = self.get_tree_id()
        if self.get_interface() is not None:
            self.get_interface().send_i_am_no_longer_upstream(source, group)

    def send_interest(self):
        """
        Send an Interest message through this interface
        """
        (source, group) = self.get_tree_id()
        best_upstream_neighbor = self._best_upstream_router.get_ip()
        if self.get_interface() is not None:
            self.get_interface().send_interest(source, group, best_upstream_neighbor)

    def send_no_interest(self):
        """
        Send a NoInterest message through this interface
        """
        (source, group) = self.get_tree_id()
        best_upstream_neighbor = self._best_upstream_router.get_ip()
        if self.get_interface() is not None:
            self.get_interface().send_no_interest(source, group, best_upstream_neighbor)

    def is_tree_active(self):
        """
        Verify if this interface considers the tree to be in Active state
        """
        return self.current_tree_state.is_active()

    def is_tree_unsure(self):
        """
        Verify if this interface considers the tree to be in Unsure state
        """
        return self.current_tree_state.is_unsure()

    def is_tree_inactive(self):
        """
        Verify if this interface considers the tree to be in Inactive state
        """
        return self.current_tree_state.is_inactive()

    #############################################################
    @abstractmethod
    def is_forwarding(self):
        """
        Determine if this interface must be included in the OIL at the multicast routing table...
        This method must be overrode by subclasses
        """
        pass

    @abstractmethod
    def delete(self):
        """
        Tree interface is being removed... due to change of interface roles or
        due to the removal of the tree by this router
        Clear all state from this interface regarding this tree
        """
        (s, g) = self.get_tree_id()
        # unsubscribe igmp information
        try:
            membership_interface = self.get_membership_interface()  # type: InterfaceIGMP
            group_state = membership_interface.interface_state.get_group_state(g)
            group_state.remove_multicast_routing_entry(self)
        except:
            pass

        print('Tree Interface deleted')

    def is_node_in_tree(self):
        """
        Determine if this router is interested in receiving data packets
        """
        return self._kernel_entry.is_in_tree()

    def evaluate_in_tree(self):
        """
        Verify if there are changes regarding interest of the router...
        This method is invoked whenever a new interface is included in the OIL or if an interface is removed from it
        """
        self._kernel_entry.evaluate_in_tree_change()

    ############################################################
    # Assert Winner state
    ############################################################
    def calculate_assert_winner(self):
        """
        Calculate the router responsible for forwarding data packets to a link...
        This method must be overrode by subclasses
        """
        return

    def change_best_upstream_neighbor_state(self, new_best_upstream_neighbor_state):
        """
        A neighbor changed Upstream state due to the reception of any control packet
        (IamUpstream or IamNoLongerUpstream or Interest or NoInterest or Sync)
        """
        self._best_upstream_router = new_best_upstream_neighbor_state

    ###########################################################
    # Interest state
    ###########################################################
    def change_interest_state(self, interest_state):
        """
        A neighbor has changed Interest state due to the reception of any control packet
        (IamUpstream or IamNoLongerUpstream or Interest or NoInterest or Sync)
        """
        return

    ##########################################################
    # Tree transitions
    ##########################################################
    def tree_transition_to_active(self):
        """
        The tree of this interface detected that the tree transitioned to Active state
        The interface must react to this change in order to send some control messages
        """
        self.current_tree_state = TreeState.Active

    def tree_transition_to_unsure(self):
        """
        The tree of this interface detected that the tree transitioned to Unsure state
        The interface must react to this change in order to send some control messages
        """
        self.current_tree_state = TreeState.Unsure

    def tree_transition_to_inactive(self):
        """
        The tree of this interface detected that the tree transitioned to Inactive state
        The interface must react to this change in order to send some control messages
        """
        self.current_tree_state = TreeState.Inactive


    #############################################################
    # Local Membership (IGMP/MLD)
    ############################################################
    def check_membership_state(self):
        """
        Reverify IGMP/MLD state of this group whenever this interface enabled or disabled IGMP/MLD
        """
        (_, group_ip) = self.get_tree_id()
        with self._igmp_lock:
            try:
                membership_interface = self.get_membership_interface()  # type: InterfaceIGMP
                group_state = membership_interface.interface_state.get_group_state(group_ip)
                has_members = group_state.add_multicast_routing_entry(self)
                self._local_membership_state = LocalMembership.Include if has_members else LocalMembership.NoInfo
            except:
                self._local_membership_state = LocalMembership.NoInfo
            self.change_tree()
            self.evaluate_in_tree()

    def notify_membership(self, has_members: bool):
        """
        IGMP/MLD detected a change of membership regarding the group of this tree
        """
        with self.get_state_lock():
            with self._igmp_lock:
                if has_members != self._local_membership_state.has_members():
                    self._local_membership_state = LocalMembership.Include if has_members else LocalMembership.NoInfo
                    self.change_tree()
                    self.evaluate_in_tree()

    def local_membership_has_members(self):
        """
        Determine if there are hosts interested in receiving data packets regarding this tree
        """
        with self._igmp_lock:
            return self._local_membership_state.has_members()

    def get_interface(self):
        """
        Get the InterfaceHPIM object regarding this physical interface
        """
        return self._kernel_entry.get_interface(self._interface_id)

    def get_membership_interface(self):
        """
        Get InterfaceIGMP or InterfaceMLD object regarding this physical interface
        """
        return self._kernel_entry.get_membership_interface(self._interface_id)

    def get_interface_name(self):
        """
        Get interface name of this interface
        """
        return self._kernel_entry.get_interface_name(self._interface_id)

    def get_ip(self):
        """
        Get IP of this interface
        """
        interface = self.get_interface()
        if interface is not None:
            return interface.get_ip()
        return "0.0.0.0"

    def get_all_interface_networks(self):
        """
        Get all networks associated with interface
        More than one can be associated for example in the case of IPv6 (link local + global + unique local)
        """
        interface = self.get_interface()
        if interface is not None:
            return interface.get_all_interface_networks()
        return {"0.0.0.0/32"}

    def is_interface_connected_to_source(self):
        """
        Determine if this interface is connected to the source of multicast traffic
        """
        source_ip = self.get_tree_id()[0]
        for network in self.get_all_interface_networks():
            if ipaddress.ip_address(source_ip) in ipaddress.ip_network(network):
                return True
        return False

    def get_tree_id(self):
        """
        Get tree id, i.e. pair (Source, Group) IPs
        """
        return (self._kernel_entry.source_ip, self._kernel_entry.group_ip)

    def change_tree(self):
        """
        Re-set multicast routing table...
        Invoked whenever there are state transitions
        """
        self._kernel_entry.change()

    def get_state_lock(self):
        """
        Obtain lock used to change state of an interface
        """
        return self._kernel_entry.CHANGE_STATE_LOCK

    ###########################################
    # Change to in/out-tree
    ###########################################
    def node_is_out_tree(self):
        return

    def node_is_in_tree(self):
        return

    ###################################################
    # RPC Change
    ###################################################
    def notify_rpc_change(self, new_rpc):
        return



import struct
from .PacketHPIMHelloOptions import PacketHPIMHelloOptionsJson, PacketHPIMHelloOptions

###########################################################################################################
# JSON FORMAT
###########################################################################################################
class PacketHPIMHelloJson:
    PIM_TYPE = "HELLO"

    def __init__(self):
        self.options = {}

    def add_option(self, option: 'PacketHPIMHelloOptionsJson'):
        self.options[option.type] = option

    def get_options(self):
        return self.options

    def bytes(self) -> dict:
        """
        Obtain Protocol Hello Packet in a format to be transmitted (JSON)
        This method will return the Hello and all its options in JSON format
        """
        res = {}
        for option in self.options.values():
            res.update(option.bytes())
        return res

    def __len__(self):
        return len(self.bytes())

    @staticmethod
    def parse_bytes(data: dict):
        """
        Parse received Hello Packet from JSON format and convert it into Hello object
        """
        pim_payload = PacketHPIMHelloJson()
        for (key, value) in data.items():
            option = PacketHPIMHelloOptionsJson.parse_bytes((key, value))
            pim_payload.add_option(option)
        return pim_payload





###########################################################################################################
# BINARY FORMAT
###########################################################################################################
'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|          Option Type          |         Option Length         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Option Value                          |
|                              ...                              |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                               .                               |
|                               .                               |
|                               .                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|          Option Type          |         Option Length         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Option Value                          |
|                              ...                              |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''

'''
NOT USED:
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Hello HoldTime                        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                          CheckpointSN                         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''


class PacketHPIMHello:
    PIM_TYPE = 0
    PIM_HDR_OPTS = "! HH"
    PIM_HDR_OPTS_LEN = struct.calcsize(PIM_HDR_OPTS)


    def __init__(self):
        self.options = {}

    def add_option(self, option: 'PacketHPIMHelloOptions'):
        self.options[option.TYPE] = option

    def get_options(self):
        return self.options

    def bytes(self) -> bytes:
        """
        Obtain Protocol Hello Packet in a format to be transmitted (binary)
        This method will return the Hello and all its options in binary format
        """
        res = b''
        for option in self.options.values():
            res += option.bytes()
        return res

    def __len__(self):
        return len(self.bytes())

    @staticmethod
    def parse_bytes(data: bytes):
        """
        Parse received Hello Packet from binary format and convert it into Hello object
        """
        protocol_payload = PacketHPIMHello()
        while data != b'':
            print("DATA", data)
            option = PacketHPIMHelloOptions.parse_bytes(data)
            option_length = len(option)
            data = data[option_length:]
            protocol_payload.add_option(option)
        return protocol_payload



from abc import ABCMeta
from hpimdm.utils import TYPE_CHECKING

if TYPE_CHECKING:
    from .KernelEntry import KernelEntry


class TreeStateABC(metaclass=ABCMeta):
    @staticmethod
    def is_active():
        """
        Verify if the interface state is Active
        """
        return False

    @staticmethod
    def is_unsure():
        """
        Verify if the interface state is Unsure
        """
        return False

    @staticmethod
    def is_inactive():
        """
        Verify if the interface state is Inactive
        """
        return False

    @staticmethod
    def transition_to_active(kernel_entry: 'KernelEntry'):
        """
        Tree transitioned to Active state
        """
        if kernel_entry.is_tree_active():
            return
        kernel_entry.set_tree_state(TreeState.Active)
        for interface in kernel_entry.interface_state.values():
            interface.tree_transition_to_active()
        kernel_entry.change()
        kernel_entry.evaluate_in_tree_change()

    @staticmethod
    def transition_to_unsure(kernel_entry: 'KernelEntry'):
        """
        Tree transitioned to Unsure state
        """
        if kernel_entry.is_tree_unsure():
            return
        kernel_entry.set_tree_state(TreeState.Unsure)
        for interface in kernel_entry.interface_state.values():
            interface.tree_transition_to_unsure()
        kernel_entry.change()
        kernel_entry.evaluate_in_tree_change()

    @staticmethod
    def transition_to_inactive(kernel_entry: 'KernelEntry'):
        """
        Tree transitioned to Inactive state
        """
        if kernel_entry.is_tree_inactive():
            return
        kernel_entry.set_tree_state(TreeState.Inactive)
        for interface in kernel_entry.interface_state.values():
            interface.tree_transition_to_inactive()
        kernel_entry.remove_entry()


class ActiveTree(TreeStateABC):
    @staticmethod
    def is_active():
        """
        Verify if the interface state is Active
        """
        return True

    def __str__(self):
        return "Active"


class UnsureTree(TreeStateABC):
    @staticmethod
    def is_unsure():
        """
        Verify if the interface state is Unsure
        """
        return True

    def __str__(self):
        return "Unsure"


class InactiveTree(TreeStateABC):
    @staticmethod
    def is_inactive():
        """
        Verify if the interface state is Inactive
        """
        return True

    def __str__(self):
        return "Inactive"


class TreeState:
    Active = ActiveTree()
    Unsure = UnsureTree()
    Inactive = InactiveTree()



from abc import ABCMeta, abstractmethod


class SFMRDownstreamStateABC(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def are_downstream_nodes_interested() -> bool:
        raise NotImplementedError()


class SFMRDownstreamInterested(SFMRDownstreamStateABC):
    @staticmethod
    def are_downstream_nodes_interested():
        """
        Determine if this state considers downstream nodes to be Interested in receiving data packets
        """
        return True

    def __str__(self):
        return 'DownstreamInterest'


class SFMRNoDownstreamInterested(SFMRDownstreamStateABC):
    @staticmethod
    def are_downstream_nodes_interested():
        """
        Determine if this state considers downstream nodes to be Interested in receiving data packets
        """
        return False

    def __str__(self):
        return 'NoDownstreamInterest'


class SFMRPruneState():
    DI = SFMRDownstreamInterested()
    NDI = SFMRNoDownstreamInterested()



import struct
import socket

###########################################################################################################
# JSON FORMAT
###########################################################################################################
class PacketHPIMInterestJson:
    PIM_TYPE = "INTEREST"

    def __init__(self, source, group, sequence_number):
        self.source = source
        self.group = group
        self.sequence_number = sequence_number

    def bytes(self) -> bytes:
        """
        Obtain Protocol Interest Packet in a format to be transmitted (JSON)
        """
        msg = {"SOURCE": self.source,
               "GROUP": self.group,
               "SN": self.sequence_number
              }

        return msg

    def __len__(self):
        return len(self.bytes())

    @classmethod
    def parse_bytes(cls, data: bytes):
        """
        Parse received Protocol Interest Packet from JSON format and convert it into ProtocolInterest object
        """
        source = data["SOURCE"]
        group = data["GROUP"]
        sn = data["SN"]
        return cls(source, group, sn)


class PacketHPIMNoInterestJson(PacketHPIMInterestJson):
    PIM_TYPE = "NO_INTEREST"

    def __init__(self, source, group, sn):
        super().__init__(source, group, sn)

###########################################################################################################
# BINARY FORMAT
###########################################################################################################
'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                        Tree Source IP                         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Tree Group IP                         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                        Sequence Number                        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''
class PacketHPIMInterest:
    PIM_TYPE = 4

    PIM_HDR_INTEREST = "! 4s 4s L"
    PIM_HDR_INTEREST_LEN = struct.calcsize(PIM_HDR_INTEREST)
    FAMILY = socket.AF_INET

    def __init__(self, source_ip, group_ip, sequence_number):
        if type(source_ip) not in (str, bytes) or type(group_ip) not in (str, bytes):
            raise Exception
        if type(source_ip) is bytes:
            source_ip = socket.inet_ntop(self.FAMILY, source_ip)
        if type(group_ip) is bytes:
            group_ip = socket.inet_ntop(self.FAMILY, group_ip)

        self.source = source_ip
        self.group = group_ip
        self.sequence_number = sequence_number

    def bytes(self) -> bytes:
        """
        Obtain Protocol Interest Packet in a format to be transmitted (binary)
        """
        msg = struct.pack(self.PIM_HDR_INTEREST, socket.inet_pton(self.FAMILY, self.source),
                          socket.inet_pton(self.FAMILY, self.group), self.sequence_number)

        return msg

    def __len__(self):
        return self.PIM_HDR_INTEREST_LEN

    @classmethod
    def parse_bytes(cls, data: bytes):
        """
        Parse received Protocol Interest Packet from binary format and convert it into ProtocolInterest object
        """
        (tree_source, tree_group, sn) = struct.unpack(cls.PIM_HDR_INTEREST, data[:cls.PIM_HDR_INTEREST_LEN])
        return cls(tree_source, tree_group, sn)


class PacketHPIMNoInterest(PacketHPIMInterest):
    PIM_TYPE = 5

    def __init__(self, source_ip, group_ip, sequence_number):
        super().__init__(source_ip, group_ip, sequence_number)


class PacketHPIMInterest_v6(PacketHPIMInterest):
    PIM_HDR_INTEREST = "! 16s 16s L"
    PIM_HDR_INTEREST_LEN = struct.calcsize(PIM_HDR_INTEREST)
    FAMILY = socket.AF_INET6

    def __init__(self, source_ip, group_ip, sequence_number):
        super().__init__(source_ip, group_ip, sequence_number)


class PacketHPIMNoInterest_v6(PacketHPIMNoInterest):
    PIM_HDR_INTEREST = "! 16s 16s L"
    PIM_HDR_INTEREST_LEN = struct.calcsize(PIM_HDR_INTEREST)
    FAMILY = socket.AF_INET6

    def __init__(self, source_ip, group_ip, sequence_number):
        super().__init__(source_ip, group_ip, sequence_number)



import socket
import struct
import traceback
import ipaddress
from threading import Thread
from socket import if_nametoindex
from abc import ABCMeta, abstractmethod

from hpimdm import Main
from hpimdm.tree import hpim_globals
from hpimdm.rwlock.RWLock import RWLockWrite
from hpimdm import UnicastRouting

from hpimdm.InterfaceMLD import InterfaceMLD
from hpimdm.InterfaceIGMP import InterfaceIGMP
from hpimdm.InterfaceHPIM import InterfaceHPIM
from hpimdm.InterfaceHPIM6 import InterfaceHPIM6
from .tree.KernelEntryInterface import KernelEntry4Interface, KernelEntry6Interface
from hpimdm.tree.KernelEntry import KernelEntry, KernelEntryOriginator, KernelEntryNonOriginator


class KernelInterface(metaclass=ABCMeta):
    MAXVIFS = 0

    def __init__(self, kernel_socket):
        # Kernel is running
        self.running = True

        # KEY : vif_index, VALUE : interface_name
        self.vif_index_to_name_dic = {}
        # KEY : interface_name, VALUE : vif_index
        self.vif_name_to_index_dic = {}

        # KEY : source_ip, VALUE : {group_ip: KernelEntry}
        self.routing = {}

        self.socket = kernel_socket
        self.rwlock = RWLockWrite()

        self.hpim_interface = {}  # name: interface_protocol
        self.membership_interface = {}  # name: interface_igmp

        # logs
        self.interface_logger = Main.logger.getChild('KernelInterface')
        self.tree_logger = Main.logger.getChild('KernelTree')

        # receive signals from kernel with a background thread
        handler_thread = Thread(target=self.handler)
        handler_thread.daemon = True
        handler_thread.start()

    ##############################################
    # Create/remove interfaces methods
    ##############################################
    @abstractmethod
    def create_virtual_interface(self, ip_interface: str or bytes, interface_name: str, index, flags=0x0):
        raise NotImplementedError

    def create_hpim_interface(self, interface_name: str):
        thread = None
        with self.rwlock.genWlock():
            hpim_interface = self.hpim_interface.get(interface_name)
            membership_interface = self.membership_interface.get(interface_name)
            vif_already_exists = hpim_interface or membership_interface
            if hpim_interface:
                # already exists
                return
            elif membership_interface:
                index = membership_interface.vif_index
            else:
                index = list(range(0, self.MAXVIFS) - self.vif_index_to_name_dic.keys())[0]

            if interface_name not in self.hpim_interface:
                hpim_interface = self._create_hpim_interface_object(interface_name, index)
                self.hpim_interface[interface_name] = hpim_interface
                ip_interface = hpim_interface.ip_interface
                if not vif_already_exists:
                    self.create_virtual_interface(ip_interface=ip_interface, interface_name=interface_name, index=index)
                hpim_interface.enable()
                thread = Thread(target=self.recheck_all_trees, args=(index,))
                thread.start()

        if thread is not None:
            thread.join()

    @abstractmethod
    def _create_hpim_interface_object(self, interface_name, index):
        raise NotImplementedError

    def create_membership_interface(self, interface_name: str):
        thread = None
        with self.rwlock.genWlock():
            hpim_interface = self.hpim_interface.get(interface_name)
            membership_interface = self.membership_interface.get(interface_name)
            vif_already_exists = hpim_interface or membership_interface
            if membership_interface:
                # already exists
                return
            elif hpim_interface:
                index = hpim_interface.vif_index
            else:
                index = list(range(0, self.MAXVIFS) - self.vif_index_to_name_dic.keys())[0]

            if interface_name not in self.membership_interface:
                membership_interface = self._create_membership_interface_object(interface_name, index)
                self.membership_interface[interface_name] = membership_interface
                ip_interface = membership_interface.ip_interface
                if not vif_already_exists:
                    self.create_virtual_interface(ip_interface=ip_interface, interface_name=interface_name, index=index)
                membership_interface.enable()
                thread = Thread(target=self.recheck_membership_all_trees, args=(index,))
                thread.start()

        if thread is not None:
            thread.join()

    @abstractmethod
    def _create_membership_interface_object(self, interface_name, index):
        raise NotImplementedError

    def remove_interface(self, interface_name, membership: bool = False, hpim: bool = False):
        thread = None
        with self.rwlock.genWlock():
            hpim_interface = self.hpim_interface.get(interface_name)
            membership_interface = self.membership_interface.get(interface_name)
            if (not membership and not hpim) or (interface_name not in self.vif_name_to_index_dic):
                return
            if hpim and hpim_interface is not None:
                hpim_interface = self.hpim_interface.pop(interface_name)
                hpim_interface.remove()
            if membership and membership_interface is not None:
                membership_interface = self.membership_interface.pop(interface_name)
                membership_interface.remove()

            if interface_name not in self.membership_interface and interface_name not in self.hpim_interface:
                self.remove_virtual_interface(interface_name)
            else:
                vif_index = self.vif_name_to_index_dic.get(interface_name)
                thread = Thread(target=self.recheck_all_trees, args=(vif_index,))
                thread.start()

        if thread is not None:
            thread.join()

    @abstractmethod
    def remove_virtual_interface(self, interface_name):
        raise NotImplementedError

    def notify_interface_changes(self, interface_name):
        with self.rwlock.genWlock():
            if interface_name is None or interface_name not in self.vif_name_to_index_dic:
                return
            igmp_was_enabled = interface_name in self.membership_interface
            hpim_was_enabled = interface_name in self.hpim_interface

        self.remove_interface(interface_name, membership=True, hpim=True)
        if igmp_was_enabled:
            self.create_membership_interface(interface_name)
        if hpim_was_enabled:
            self.create_hpim_interface(interface_name)

    def add_interface_security(self, interface_name, security_id, security_algorithm, security_key):
        with self.rwlock.genRlock():
            if interface_name not in self.hpim_interface:
                return

            interface = self.hpim_interface.get(interface_name) #type: InterfaceHPIM
            interface.add_security_key(security_id, security_algorithm, security_key)

    def remove_interface_security(self, interface_name, security_id):
        with self.rwlock.genRlock():
            if interface_name not in self.hpim_interface:
                return

            interface = self.hpim_interface.get(interface_name) #type: InterfaceHPIM
            interface.remove_security_key(security_id)

    #############################################
    # Manipulate multicast routing table
    #############################################
    @abstractmethod
    def set_multicast_route(self, kernel_entry: KernelEntry):
        raise NotImplementedError

    @abstractmethod
    def set_flood_multicast_route(self, source_ip, group_ip, inbound_interface_index):
        raise NotImplementedError

    def remove_multicast_route(self, kernel_entry: KernelEntry):
        Thread(target=self._remove_multicast_route, args=(kernel_entry,)).start()
        #self._remove_multicast_route(kernel_entry, struct_mfcctl)

    @abstractmethod
    def _remove_multicast_route(self, kernel_entry):
        raise NotImplementedError

    @abstractmethod
    def exit(self):
        raise NotImplementedError

    @abstractmethod
    def handler(self):
        raise NotImplementedError

    def notify_unicast_changes(self, subnet):
        # notify KernelEntries about changes at the unicast routing table
        with self.rwlock.genWlock():
            for source_ip in list(self.routing.keys()):
                source_ip_obj = ipaddress.ip_address(source_ip)
                if source_ip_obj not in subnet:
                    continue
                for group_ip in list(self.routing[source_ip].keys()):
                    self.routing[source_ip][group_ip].network_update()

    ############################################################
    # Interact with received control packets
    ############################################################
    def recv_upstream_msg(self, source_group, interface: "InterfaceHPIM"):
        ip_src = source_group[0]
        ip_dst = source_group[1]

        print("ENTER RCV_UPSTREAM")
        with self.rwlock.genWlock():
            if interface not in self.hpim_interface.values():
                return

            (interest_state, upstream_state) = interface.get_tree_state(source_group)
            tree_is_not_inactive = upstream_state is not None
            print("RCV INSTALL/UNINSTALL")
            print("INTEREST: ", interest_state)

            if tree_is_not_inactive and (ip_src not in self.routing or ip_dst not in self.routing.get(ip_src, {})):
                self.create_entry(ip_src, ip_dst)
            elif ip_src in self.routing and ip_dst in self.routing[ip_src]:
                self.routing[ip_src][ip_dst].check_interface_state(interface.vif_index, upstream_state, interest_state)
            else:
                interface.remove_tree_state(ip_src, ip_dst)
        print("EXIT RCV_UPSTREAM")

    def recv_interest_msg(self, source_group, interface: "InterfaceHPIM"):
        ip_src = source_group[0]
        ip_dst = source_group[1]

        print("ENTER RECV_INTEREST")
        with self.rwlock.genRlock():
            if interface not in self.hpim_interface.values():
                return

            if ip_src not in self.routing or ip_dst not in self.routing[ip_src]:
                interface.remove_tree_state(ip_src, ip_dst)
                return

            (interest_state, upstream_state) = interface.get_tree_state(source_group)
            self.routing[ip_src][ip_dst].check_interest_state(interface.vif_index, interest_state)
        print("EXIT RECV_INTEREST")

    #############################################################
    # Create kernel entries (data structure representing a tree)
    #############################################################
    def create_entry(self, ip_src, ip_dst):
        (_, _, is_directly_connected, _) = UnicastRouting.get_unicast_info(ip_src)

        upstream_state_dict = {}
        interest_state_dict = {}
        for interface in self.hpim_interface.values():
            (interest_state, upstream_state) = interface.get_tree_state((ip_src, ip_dst))
            upstream_state_dict[interface.vif_index] = upstream_state
            interest_state_dict[interface.vif_index] = interest_state

        other_interfaces = self.hpim_interface.keys() - self.vif_name_to_index_dic.keys()
        for interface_name in other_interfaces:
            vif_index = self.vif_name_to_index_dic.get(interface_name)
            upstream_state_dict[vif_index] = None
            interest_state_dict[vif_index] = False

        if ip_src not in self.routing:
            self.routing[ip_src] = {}

        if ip_dst not in self.routing[ip_src] and is_directly_connected:
            self.routing[ip_src][ip_dst] = KernelEntryOriginator(ip_src, ip_dst,
                                                                 upstream_state_dict, interest_state_dict,
                                                                 self._get_kernel_entry_interface())
        elif ip_dst not in self.routing[ip_src]:
            self.routing[ip_src][ip_dst] = KernelEntryNonOriginator(ip_src, ip_dst,
                                                                    upstream_state_dict, interest_state_dict,
                                                                    self._get_kernel_entry_interface())

    @staticmethod
    @abstractmethod
    def _get_kernel_entry_interface():
        pass

    def snapshot_multicast_routing_table(self, vif_index):
        trees_to_sync = {}
        print("ENTER SNAPSHOT")
        #with self.rwlock.genWlock():
        for (ip_src, src_dict) in self.routing.items():
            for (ip_dst, kernel_entry) in self.routing[ip_src].items():
                tree = kernel_entry.get_interface_sync_state(vif_index)
                if tree is not None:
                    trees_to_sync[(ip_src, ip_dst)] = tree
        print("EXIT SNAPSHOT")
        return trees_to_sync

    def recheck_all_trees(self, vif_index: int):
        print("ENTER RECHECK")
        with self.rwlock.genWlock():
            interface_name = self.vif_index_to_name_dic.get(vif_index, None)
            interface = self.hpim_interface.get(interface_name, None)

            known_trees = set()
            if interface is not None:
                for n in list(interface.neighbors.values()):
                    known_trees = known_trees.union(n.get_known_trees())

            for (source, src_dict) in self.routing.items():
                for group in src_dict.keys():
                    known_trees.add((source, group))

            print("KNOWN TREES: ", known_trees)
            #print("INDEX:", interface.vif_index)

            for tree in known_trees:
                if interface is not None:
                    (interest_state, upstream_state) = interface.get_tree_state(tree)
                else:
                    (interest_state, upstream_state) = (False, None)

                print("UPSTREAM_STATE:", upstream_state)
                print("INTEREST_STATE:", interest_state)
                if upstream_state is not None and (tree[0] not in self.routing or tree[1] not in self.routing.get(tree[0], {})):
                    self.create_entry(tree[0], tree[1])
                elif tree[0] in self.routing and tree[1] in self.routing[tree[0]]:
                    self.routing[tree[0]][tree[1]].check_interface_state(vif_index, upstream_state, interest_state)
        print("EXIT RECHECK")

    def recheck_membership_all_trees(self, vif_index: int):
        print("ENTER RECHECK IGMP")
        with self.rwlock.genWlock():
            for src_dict in self.routing.values():
                for entry in src_dict.values():
                    entry.check_membership_state(vif_index)
            print("EXIT RECHECK IGMP")

    def recheck_all_trees_in_all_interfaces(self):
        for i in list(self.vif_index_to_name_dic.keys()):
            self.recheck_all_trees(i)


class Kernel4(KernelInterface):
    # MRT
    MRT_BASE    = 200
    MRT_INIT    = (MRT_BASE)      # /* Activate the kernel mroute code 	*/
    MRT_DONE    = (MRT_BASE + 1)  # /* Shutdown the kernel mroute		*/
    MRT_ADD_VIF = (MRT_BASE + 2)  # /* Add a virtual interface		    */
    MRT_DEL_VIF = (MRT_BASE + 3)  # /* Delete a virtual interface		*/
    MRT_ADD_MFC = (MRT_BASE + 4)  # /* Add a multicast forwarding entry	*/
    MRT_DEL_MFC = (MRT_BASE + 5)  # /* Delete a multicast forwarding entry	*/
    MRT_VERSION = (MRT_BASE + 6)  # /* Get the kernel multicast version	*/
    MRT_ASSERT  = (MRT_BASE + 7)  # /* Activate PIM assert mode		    */
    MRT_PIM     = (MRT_BASE + 8)  # /* enable PIM code			        */
    MRT_TABLE   = (MRT_BASE + 9)  # /* Specify mroute table ID		    */
    #MRT_ADD_MFC_PROXY = (MRT_BASE + 10)  # /* Add a (*,*|G) mfc entry	*/
    #MRT_DEL_MFC_PROXY = (MRT_BASE + 11)  # /* Del a (*,*|G) mfc entry	*/
    #MRT_MAX = (MRT_BASE + 11)

    # Max Number of Virtual Interfaces
    MAXVIFS = 32

    # SIGNAL MSG TYPE
    IGMPMSG_NOCACHE = 1
    IGMPMSG_WRONGVIF = 2
    IGMPMSG_WHOLEPKT = 3  # NOT USED ON PIM-DM

    # Interface flags
    VIFF_TUNNEL      = 0x1  # IPIP tunnel
    VIFF_SRCRT       = 0x2  # NI
    VIFF_REGISTER    = 0x4  # register vif
    VIFF_USE_IFINDEX = 0x8  # use vifc_lcl_ifindex instead of vifc_lcl_addr to find an interface

    def __init__(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_IGMP)

        # MRT TABLE
        if hpim_globals.MULTICAST_TABLE_ID != 0:
            try:
                s.setsockopt(socket.IPPROTO_IP, Kernel4.MRT_TABLE, hpim_globals.MULTICAST_TABLE_ID)
            except:
                traceback.print_exc()

        # MRT INIT
        s.setsockopt(socket.IPPROTO_IP, Kernel4.MRT_INIT, 1)

        # MRT PIM
        s.setsockopt(socket.IPPROTO_IP, Kernel4.MRT_PIM, 0)
        s.setsockopt(socket.IPPROTO_IP, Kernel4.MRT_ASSERT, 0)

        super().__init__(s)

    '''
    Structure to create/remove virtual interfaces
    struct vifctl {
        vifi_t	vifc_vifi;		        /* Index of VIF */
        unsigned char vifc_flags;	    /* VIFF_ flags */
        unsigned char vifc_threshold;	/* ttl limit */
        unsigned int vifc_rate_limit;	/* Rate limiter values (NI) */
        union {
            struct in_addr vifc_lcl_addr;     /* Local interface address */
            int            vifc_lcl_ifindex;  /* Local interface index   */
        };
        struct in_addr vifc_rmt_addr;	/* IPIP tunnel addr */
    };
    '''
    def create_virtual_interface(self, ip_interface: str or bytes, interface_name: str, index, flags=0x0):
        if type(ip_interface) is str:
            ip_interface = socket.inet_aton(ip_interface)

        struct_mrt_add_vif = struct.pack("HBBI 4s 4s", index, flags, 1, 0, ip_interface,
                                         socket.inet_aton("0.0.0.0"))
        self.socket.setsockopt(socket.IPPROTO_IP, Kernel4.MRT_ADD_VIF, struct_mrt_add_vif)
        self.vif_index_to_name_dic[index] = interface_name
        self.vif_name_to_index_dic[interface_name] = index

        for source_dict in list(self.routing.values()):
            for kernel_entry in list(source_dict.values()):
                kernel_entry.new_interface(index)

        self.interface_logger.debug('Create virtual interface: %s -> %d', interface_name, index)
        return index

    def _create_hpim_interface_object(self, interface_name, index):
        return InterfaceHPIM(interface_name, index)

    def _create_membership_interface_object(self, interface_name, index):
        return InterfaceIGMP(interface_name, index)

    def remove_virtual_interface(self, interface_name):
        index = self.vif_name_to_index_dic.pop(interface_name, None)
        struct_vifctl = struct.pack("HBBI 4s 4s", index, 0, 0, 0, socket.inet_aton("0.0.0.0"),
                                    socket.inet_aton("0.0.0.0"))

        try:
            self.socket.setsockopt(socket.IPPROTO_IP, Kernel4.MRT_DEL_VIF, struct_vifctl)
        except socket.error:
            pass

        self.vif_index_to_name_dic.pop(index, None)

        # remove this interface from KernelEntries
        for source_dict in list(self.routing.values()):
            for kernel_entry in list(source_dict.values()):
                kernel_entry.remove_interface(index)

        self.interface_logger.debug('Remove virtual interface: %s -> %d', interface_name, index)

    '''
    /* Cache manipulation structures for mrouted and PIMd */
    struct mfcctl {
        struct in_addr mfcc_origin;		    /* Origin of mcast	*/
        struct in_addr mfcc_mcastgrp;		/* Group in question	*/
        vifi_t mfcc_parent; 			    /* Where it arrived	*/
        unsigned char mfcc_ttls[MAXVIFS];	/* Where it is going	*/
        unsigned int mfcc_pkt_cnt;		    /* pkt count for src-grp */
        unsigned int mfcc_byte_cnt;
        unsigned int mfcc_wrong_if;
        int mfcc_expire;
    };
    '''
    def set_multicast_route(self, kernel_entry: KernelEntry):
        source_ip = socket.inet_aton(kernel_entry.source_ip)
        group_ip = socket.inet_aton(kernel_entry.group_ip)

        outbound_interfaces = kernel_entry.get_outbound_interfaces_indexes()
        if len(outbound_interfaces) != Kernel4.MAXVIFS:
            raise Exception

        #outbound_interfaces_and_other_parameters = list(kernel_entry.outbound_interfaces) + [0]*4
        outbound_interfaces_and_other_parameters = outbound_interfaces + [0]*4

        #outbound_interfaces, 0, 0, 0, 0 <- only works with python>=3.5
        #struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces, 0, 0, 0, 0)
        struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel4.MAXVIFS + " IIIi", source_ip, group_ip, kernel_entry.inbound_interface_index, *outbound_interfaces_and_other_parameters)
        self.socket.setsockopt(socket.IPPROTO_IP, Kernel4.MRT_ADD_MFC, struct_mfcctl)

    def set_flood_multicast_route(self, source_ip, group_ip, inbound_interface_index):
        if inbound_interface_index is None:
            return
        source_ip = socket.inet_aton(source_ip)
        group_ip = socket.inet_aton(group_ip)

        outbound_interfaces = [1] * self.MAXVIFS
        outbound_interfaces[inbound_interface_index] = 0

        # outbound_interfaces_and_other_parameters = list(kernel_entry.outbound_interfaces) + [0]*4
        outbound_interfaces_and_other_parameters = outbound_interfaces + [0] * 3 + [hpim_globals.INITIAL_FLOOD_TIME]

        # outbound_interfaces, 0, 0, 0, 0 <- only works with python>=3.5
        # struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces, 0, 0, 0, 0)
        struct_mfcctl = struct.pack("4s 4s H " + "B" * Kernel4.MAXVIFS + " IIIi", source_ip, group_ip,
                                    inbound_interface_index, *outbound_interfaces_and_other_parameters)
        self.socket.setsockopt(socket.IPPROTO_IP, Kernel4.MRT_ADD_MFC, struct_mfcctl)

    def _remove_multicast_route(self, kernel_entry):
        source_ip = socket.inet_aton(kernel_entry.source_ip)
        group_ip = socket.inet_aton(kernel_entry.group_ip)
        outbound_interfaces_and_other_parameters = [0] + [0] * Kernel4.MAXVIFS + [0] * 4

        struct_mfcctl = struct.pack("4s 4s H " + "B" * Kernel4.MAXVIFS + " IIIi", source_ip, group_ip,
                                    *outbound_interfaces_and_other_parameters)
        with self.rwlock.genWlock():
            source_ip = kernel_entry.source_ip
            group_ip = kernel_entry.group_ip

            if not (source_ip in self.routing and group_ip in self.routing[source_ip]):
                return
            try:
                self.socket.setsockopt(socket.IPPROTO_IP, Kernel4.MRT_DEL_MFC, struct_mfcctl)
            except socket.error:
                pass
            self.routing[kernel_entry.source_ip].pop(kernel_entry.group_ip)
            kernel_entry.delete_state()
            if len(self.routing[source_ip]) == 0:
                self.routing.pop(source_ip)

            info_about_tree_can_be_removed = True
            for interface in self.hpim_interface.values():
                if not info_about_tree_can_be_removed:
                    break
                for n in list(interface.neighbors.values()):
                    (_, neighbor_upstream_state) = n.get_tree_state(tree=(source_ip, group_ip))
                    if neighbor_upstream_state is not None:
                        self.create_entry(kernel_entry.source_ip, kernel_entry.group_ip)
                        info_about_tree_can_be_removed = False
                        break

            if info_about_tree_can_be_removed:
                for interface in self.hpim_interface.values():
                    interface.remove_tree_state(kernel_entry.source_ip, kernel_entry.group_ip)

    def exit(self):
        self.running = False

        # MRT DONE
        self.socket.setsockopt(socket.IPPROTO_IP, Kernel4.MRT_DONE, 1)
        self.socket.close()

    '''
    /* This is the format the mroute daemon expects to see IGMP control
    * data. Magically happens to be like an IP packet as per the original
    */
    struct igmpmsg {
        __u32 unused1,unused2;
        unsigned char im_msgtype;		/* What is this */
        unsigned char im_mbz;			/* Must be zero */
        unsigned char im_vif;			/* Interface (this ought to be a vifi_t!) */
        unsigned char unused3;
        struct in_addr im_src,im_dst;
    };
    '''
    def handler(self):
        while self.running:
            try:
                msg = self.socket.recv(20)
                (_, _, im_msgtype, im_mbz, im_vif, _, im_src, im_dst) = struct.unpack("II B B B B 4s 4s", msg[:20])
                print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                if im_mbz != 0:
                    continue

                print(im_msgtype)
                print(im_mbz)
                print(im_vif)
                print(socket.inet_ntoa(im_src))
                print(socket.inet_ntoa(im_dst))
                #print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                ip_src = socket.inet_ntoa(im_src)
                ip_dst = socket.inet_ntoa(im_dst)

                if im_msgtype == Kernel4.IGMPMSG_NOCACHE:
                    print("IGMP NO CACHE")
                    self.igmpmsg_nocache_handler(ip_src, ip_dst, im_vif)
                elif im_msgtype == Kernel4.IGMPMSG_WRONGVIF:
                    print("WRONG VIF HANDLER")
                    self.igmpmsg_wrongvif_handler(ip_src, ip_dst, im_vif)
                #elif im_msgtype == Kernel.IGMPMSG_WHOLEPKT:
                #    print("IGMP_WHOLEPKT")
                #    self.igmpmsg_wholepacket_handler(ip_src, ip_dst)
                else:
                    raise Exception
            except Exception:
                traceback.print_exc()
                continue

    # receive multicast (S,G) packet and multicast routing table has no (S,G) entry
    def igmpmsg_nocache_handler(self, ip_src, ip_dst, iif):
        (_, _, is_directly_connected, rpf_if) = UnicastRouting.get_unicast_info(ip_src)

        with self.rwlock.genWlock():
            if ip_src in self.routing and ip_dst in self.routing[ip_src]:
                self.routing[ip_src][ip_dst].recv_data_msg(iif)
            elif is_directly_connected:
                if hpim_globals.INITIAL_FLOOD_ENABLED:
                    # flood
                    self.set_flood_multicast_route(ip_src, ip_dst, rpf_if)
                if rpf_if is not None:
                    self.create_entry(ip_src, ip_dst)
                    self.routing[ip_src][ip_dst].recv_data_msg(iif)
            elif not is_directly_connected and hpim_globals.INITIAL_FLOOD_ENABLED:
                # flood
                self.set_flood_multicast_route(ip_src, ip_dst, rpf_if)

    # receive multicast (S,G) packet in a outbound_interface
    def igmpmsg_wrongvif_handler(self, ip_src, ip_dst, iif):
        #source_group_pair = (ip_src, ip_dst)
        #self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg(iif)
        return

    ''' useless in PIM-DM... useful in PIM-SM
    def igmpmsg_wholepacket_handler(self, ip_src, ip_dst):
        #kernel_entry = self.routing[(ip_src, ip_dst)]
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg()
        #kernel_entry.recv_data_msg(iif)
    '''

    @staticmethod
    def _get_kernel_entry_interface():
        return KernelEntry4Interface


class Kernel6(KernelInterface):
    # MRT6
    MRT6_BASE           = 200
    MRT6_INIT           = (MRT6_BASE)       #/* Activate the kernel mroute code 	*/
    MRT6_DONE           = (MRT6_BASE+1)     #/* Shutdown the kernel mroute		*/
    MRT6_ADD_MIF        = (MRT6_BASE+2)	    #/* Add a virtual interface		*/
    MRT6_DEL_MIF        = (MRT6_BASE+3)	    #/* Delete a virtual interface		*/
    MRT6_ADD_MFC        = (MRT6_BASE+4)	    #/* Add a multicast forwarding entry	*/
    MRT6_DEL_MFC        = (MRT6_BASE+5)	    #/* Delete a multicast forwarding entry	*/
    MRT6_VERSION        = (MRT6_BASE+6)	    #/* Get the kernel multicast version	*/
    MRT6_ASSERT         = (MRT6_BASE+7)     #/* Activate PIM assert mode		*/
    MRT6_PIM            = (MRT6_BASE+8)     #/* enable PIM code			*/
    MRT6_TABLE          = (MRT6_BASE+9)	    #/* Specify mroute table ID		*/
    MRT6_ADD_MFC_PROXY  = (MRT6_BASE+10)    #/* Add a (*,*|G) mfc entry	*/
    MRT6_DEL_MFC_PROXY  = (MRT6_BASE+11)    #/* Del a (*,*|G) mfc entry	*/
    MRT6_MAX            = (MRT6_BASE+11)

    # define SIOCGETMIFCNT_IN6	SIOCPROTOPRIVATE	/* IP protocol privates */
    # define SIOCGETSGCNT_IN6	(SIOCPROTOPRIVATE+1)
    # define SIOCGETRPF	(SIOCPROTOPRIVATE+2)

    # Max Number of Virtual Interfaces
    MAXVIFS	 = 32

    # SIGNAL MSG TYPE
    MRT6MSG_NOCACHE  = 1
    MRT6MSG_WRONGMIF = 2
    MRT6MSG_WHOLEPKT = 3		# /* used for use level encap */

    # Interface flags
    MIFF_REGISTER = 0x1         # /* register vif	*/

    def __init__(self):
        s = socket.socket(socket.AF_INET6, socket.SOCK_RAW, socket.IPPROTO_ICMPV6)

        # MRT TABLE
        if hpim_globals.MULTICAST_TABLE_ID != 0:
            try:
                s.setsockopt(socket.IPPROTO_IPV6, Kernel6.MRT6_TABLE, hpim_globals.MULTICAST_TABLE_ID)
            except:
                traceback.print_exc()

        # MRT INIT
        s.setsockopt(socket.IPPROTO_IPV6, Kernel6.MRT6_INIT, 1)

        # MRT PIM
        s.setsockopt(socket.IPPROTO_IPV6, Kernel6.MRT6_PIM, 0)
        s.setsockopt(socket.IPPROTO_IPV6, Kernel6.MRT6_ASSERT, 0)

        super().__init__(s)

    '''
    Structure to create/remove multicast interfaces
    struct mif6ctl {
        mifi_t	mif6c_mifi;		        /* Index of MIF */
        unsigned char mif6c_flags;	    /* MIFF_ flags */
        unsigned char vifc_threshold;	/* ttl limit */
        __u16	 mif6c_pifi;		    /* the index of the physical IF */
        unsigned int vifc_rate_limit;	/* Rate limiter values (NI) */
    };
    '''
    def create_virtual_interface(self, ip_interface, interface_name: str, index, flags=0x0):
        physical_if_index = if_nametoindex(interface_name)
        struct_mrt_add_vif = struct.pack("HBBHI", index, flags, 1, physical_if_index, 0)
        self.socket.setsockopt(socket.IPPROTO_IPV6, Kernel6.MRT6_ADD_MIF, struct_mrt_add_vif)
        self.vif_index_to_name_dic[index] = interface_name
        self.vif_name_to_index_dic[interface_name] = index

        for source_dict in list(self.routing.values()):
            for kernel_entry in list(source_dict.values()):
                kernel_entry.new_interface(index)

        self.interface_logger.debug('Create virtual interface: %s -> %d', interface_name, index)
        return index

    def _create_hpim_interface_object(self, interface_name, index):
        return InterfaceHPIM6(interface_name, index)

    def _create_membership_interface_object(self, interface_name, index):
        return InterfaceMLD(interface_name, index)

    def remove_virtual_interface(self, interface_name):
        mif_index = self.vif_name_to_index_dic.pop(interface_name, None)
        interface_name = self.vif_index_to_name_dic.pop(mif_index)

        physical_if_index = if_nametoindex(interface_name)
        struct_vifctl = struct.pack("HBBHI", mif_index, 0, 0, physical_if_index, 0)

        try:
            self.socket.setsockopt(socket.IPPROTO_IPV6, Kernel6.MRT6_DEL_MIF, struct_vifctl)
        except socket.error:
            pass

        # remove this interface from KernelEntries
        for source_dict in list(self.routing.values()):
            for kernel_entry in list(source_dict.values()):
                kernel_entry.remove_interface(mif_index)

        self.interface_logger.debug('Remove virtual interface: %s -> %d', interface_name, mif_index)

    '''
    /* Cache manipulation structures for mrouted and PIMd */
    typedef	__u32 if_mask;
    typedef struct if_set {
        if_mask ifs_bits[__KERNEL_DIV_ROUND_UP(IF_SETSIZE, NIFBITS)];
    } if_set;

    struct mf6cctl {
        struct sockaddr_in6 mf6cc_origin;		/* Origin of mcast	*/
        struct sockaddr_in6 mf6cc_mcastgrp;		/* Group in question	*/
        mifi_t	mf6cc_parent;			        /* Where it arrived	*/
        struct if_set mf6cc_ifset;		        /* Where it is going */
    };
    '''
    def set_multicast_route(self, kernel_entry: KernelEntry):
        source_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.source_ip)
        sockaddr_in6_source = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, source_ip, 0)
        group_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.group_ip)
        sockaddr_in6_group = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, group_ip, 0)

        outbound_interfaces = kernel_entry.get_outbound_interfaces_indexes()
        if len(outbound_interfaces) != 8:
            raise Exception

        struct_mf6cctl = struct.pack("28s 28s H " + "I" * 8, sockaddr_in6_source, sockaddr_in6_group,
                                     kernel_entry.inbound_interface_index,
                                     *outbound_interfaces)
        self.socket.setsockopt(socket.IPPROTO_IPV6, Kernel6.MRT6_ADD_MFC, struct_mf6cctl)

    def set_flood_multicast_route(self, source_ip, group_ip, inbound_interface_index):
        source_ip = socket.inet_pton(socket.AF_INET6, source_ip)
        sockaddr_in6_source = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, source_ip, 0)
        group_ip = socket.inet_pton(socket.AF_INET6, group_ip)
        sockaddr_in6_group = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, group_ip, 0)

        outbound_interfaces = [255] * 8
        outbound_interfaces[inbound_interface_index // 32] = 0xFFFFFFFF & ~(1 << (inbound_interface_index % 32))

        struct_mf6cctl = struct.pack("28s 28s H " + "I" * 8, sockaddr_in6_source, sockaddr_in6_group,
                                     inbound_interface_index, *outbound_interfaces)
        self.socket.setsockopt(socket.IPPROTO_IPV6, Kernel6.MRT6_ADD_MFC, struct_mf6cctl)

    def _remove_multicast_route(self, kernel_entry):
        source_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.source_ip)
        sockaddr_in6_source = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, source_ip, 0)
        group_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.group_ip)
        sockaddr_in6_group = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, group_ip, 0)
        outbound_interfaces = [0] * 8

        struct_mf6cctl = struct.pack("28s 28s H " + "I" * 8, sockaddr_in6_source, sockaddr_in6_group, 0,
                                     *outbound_interfaces)
        with self.rwlock.genWlock():
            source_ip = kernel_entry.source_ip
            group_ip = kernel_entry.group_ip

            if not (source_ip in self.routing and group_ip in self.routing[source_ip]):
                return
            try:
                self.socket.setsockopt(socket.IPPROTO_IPV6, Kernel6.MRT6_DEL_MFC, struct_mf6cctl)
            except socket.error:
                pass
            self.routing[kernel_entry.source_ip].pop(kernel_entry.group_ip)
            kernel_entry.delete_state()
            if len(self.routing[source_ip]) == 0:
                self.routing.pop(source_ip)

            info_about_tree_can_be_removed = True
            for interface in self.hpim_interface.values():
                if not info_about_tree_can_be_removed:
                    break
                for n in list(interface.neighbors.values()):
                    (_, neighbor_upstream_state) = n.get_tree_state(tree=(source_ip, group_ip))
                    if neighbor_upstream_state is not None:
                        self.create_entry(kernel_entry.source_ip, kernel_entry.group_ip)
                        info_about_tree_can_be_removed = False
                        break

            if info_about_tree_can_be_removed:
                for interface in self.hpim_interface.values():
                    interface.remove_tree_state(kernel_entry.source_ip, kernel_entry.group_ip)

    def exit(self):
        self.running = False

        # MRT DONE
        self.socket.setsockopt(socket.IPPROTO_IPV6, Kernel6.MRT6_DONE, 1)
        self.socket.close()

    '''
    /*
     * Structure used to communicate from kernel to multicast router.
     * We'll overlay the structure onto an MLD header (not an IPv6 heder like igmpmsg{}
     * used for IPv4 implementation). This is because this structure will be passed via an
     * IPv6 raw socket, on which an application will only receiver the payload i.e the data after
     * the IPv6 header and all the extension headers. (See section 3 of RFC 3542)
     */

    struct mrt6msg {
        __u8		im6_mbz;		/* must be zero		   */
        __u8		im6_msgtype;	/* what type of message    */
        __u16		im6_mif;		/* mif rec'd on		   */
        __u32		im6_pad;		/* padding for 64 bit arch */
        struct in6_addr	im6_src, im6_dst;
    };

    /* ip6mr netlink cache report attributes */
    enum {
        IP6MRA_CREPORT_UNSPEC,
        IP6MRA_CREPORT_MSGTYPE,
        IP6MRA_CREPORT_MIF_ID,
        IP6MRA_CREPORT_SRC_ADDR,
        IP6MRA_CREPORT_DST_ADDR,
        IP6MRA_CREPORT_PKT,
        __IP6MRA_CREPORT_MAX
    };
    '''
    def handler(self):
        while self.running:
            try:
                msg = self.socket.recv(500)
                if len(msg) < 40:
                    continue
                (im6_mbz, im6_msgtype, im6_mif, _, im6_src, im6_dst) = struct.unpack("B B H I 16s 16s", msg[:40])
                # print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                if im6_mbz != 0:
                    continue

                print(im6_mbz)
                print(im6_msgtype)
                print(im6_mif)
                print(socket.inet_ntop(socket.AF_INET6, im6_src))
                print(socket.inet_ntop(socket.AF_INET6, im6_dst))
                # print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                ip_src = socket.inet_ntop(socket.AF_INET6, im6_src)
                ip_dst = socket.inet_ntop(socket.AF_INET6, im6_dst)

                if im6_msgtype == Kernel6.MRT6MSG_NOCACHE:
                    print("MRT6 NO CACHE")
                    self.msg_nocache_handler(ip_src, ip_dst, im6_mif)
                elif im6_msgtype == Kernel6.MRT6MSG_WRONGMIF:
                    print("WRONG MIF HANDLER")
                    self.msg_wrongvif_handler(ip_src, ip_dst, im6_mif)
                # elif im_msgtype == Kernel.IGMPMSG_WHOLEPKT:
                #    print("IGMP_WHOLEPKT")
                #    self.igmpmsg_wholepacket_handler(ip_src, ip_dst)
                else:
                    raise Exception
            except Exception:
                traceback.print_exc()
                continue

    # receive multicast (S,G) packet and multicast routing table has no (S,G) entry
    def msg_nocache_handler(self, ip_src, ip_dst, iif):
        (_, _, is_directly_connected, rpf_if) = UnicastRouting.get_unicast_info(ip_src)

        with self.rwlock.genWlock():
            if ip_src in self.routing and ip_dst in self.routing[ip_src]:
                self.routing[ip_src][ip_dst].recv_data_msg(iif)
            elif is_directly_connected:
                if hpim_globals.INITIAL_FLOOD_ENABLED:
                    # flood
                    self.set_flood_multicast_route(ip_src, ip_dst, rpf_if)
                if rpf_if is not None:
                    self.create_entry(ip_src, ip_dst)
                    self.routing[ip_src][ip_dst].recv_data_msg(iif)
            elif not is_directly_connected and hpim_globals.INITIAL_FLOOD_ENABLED:
                # flood
                self.set_flood_multicast_route(ip_src, ip_dst, rpf_if)

    # receive multicast (S,G) packet in a outbound_interface
    def msg_wrongvif_handler(self, ip_src, ip_dst, iif):
        #source_group_pair = (ip_src, ip_dst)
        #self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg(iif)
        return

    @staticmethod
    def _get_kernel_entry_interface():
        return KernelEntry6Interface



import ipaddress


class Metric(object):
    def __init__(self, metric_preference: int = 0x7FFFFFFF, route_metric: int = 0xFFFFFFFF):
        self._metric_preference = metric_preference
        self._route_metric = route_metric

    def is_better_than(self, other):
        """
        Verify if this RPC state is better than other
        """
        if self.metric_preference != other.metric_preference:
            return self.metric_preference < other.metric_preference
        else:
            return self.route_metric < other.route_metric

    def __eq__(self, other):
        return self.metric_preference == other.metric_preference and self.route_metric == other.route_metric

    @property
    def metric_preference(self):
        """
        Obtain metric preference of this Metric
        """
        return self._metric_preference

    @metric_preference.setter
    def metric_preference(self, value):
        """
        Set metric preference of this Metric
        """
        self._metric_preference = value

    @property
    def route_metric(self):
        """
        Obtain route metric of this Metric
        """
        return self._route_metric

    @route_metric.setter
    def route_metric(self, value):
        """
        Set route metric of this Metric
        """
        self._route_metric = value


class AssertMetric(Metric):
    def __init__(self, metric_preference: int = 0x7FFFFFFF, route_metric: int = 0xFFFFFFFF, ip_address: str = "0.0.0.0"):
        super().__init__(metric_preference, route_metric)

        if type(ip_address) is str:
            ip_address = ipaddress.ip_address(ip_address)

        self._ip_address = ip_address

    def is_better_than(self, other):
        """
        Verify if this RPC state is better than other
        """
        return super().is_better_than(other) or (super().__eq__(other) and self.ip_address > other.ip_address)

    def is_worse(self, other):
        """
        Verify if this RPC state is worse than other
        """
        return not self.is_better_than(other)

    def i_am_assert_winner(self, tree_if):
        """
        Verify if this AssertMetric is storing my state
        """
        return self.get_ip() == tree_if.get_ip()

    @property
    def ip_address(self):
        """
        Obtain IP address of this AssertMetric
        """
        return self._ip_address

    @ip_address.setter
    def ip_address(self, value):
        """
        Set IP address of this AssertMetric
        """
        if type(value) is str:
            value = ipaddress.ip_address(value)

        self._ip_address = value

    def get_ip(self):
        """
        Obtain IP address of this AssertMetric in a string format
        """
        return str(self._ip_address)

    def __str__(self):
        return "Metric Preference: " + str(self.metric_preference) + "; Metric: " + str(self.route_metric)



#!/usr/bin/env python3
"""
Read Write Lock
"""

import threading
import time


class RWLockRead(object):
    """
    A Read/Write lock giving preference to Reader
    """
    def __init__(self):
        self.V_ReadCount = 0
        self.A_Resource = threading.Lock()
        self.A_LockReadCount = threading.Lock()

    class _aReader(object):
        def __init__(self, p_RWLock):
            self.A_RWLock = p_RWLock
            self.V_Locked = False

        def acquire(self, blocking=1, timeout=-1):
            p_TimeOut = None if (blocking and timeout < 0) else (timeout if blocking else 0)
            c_DeadLine = None if p_TimeOut is None else (time.time() + p_TimeOut)
            if not self.A_RWLock.A_LockReadCount.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
                return False
            self.A_RWLock.V_ReadCount += 1
            if self.A_RWLock.V_ReadCount == 1:
                if not self.A_RWLock.A_Resource.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
                    self.A_RWLock.V_ReadCount -= 1
                    self.A_RWLock.A_LockReadCount.release()
                    return False
            self.A_RWLock.A_LockReadCount.release()
            self.V_Locked = True
            return True

        def release(self):
            if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
            self.V_Locked = False
            self.A_RWLock.A_LockReadCount.acquire()
            self.A_RWLock.V_ReadCount -= 1
            if self.A_RWLock.V_ReadCount == 0:
                self.A_RWLock.A_Resource.release()
            self.A_RWLock.A_LockReadCount.release()

        def locked(self):
            return self.V_Locked

        def __enter__(self):
            self.acquire()

        def __exit__(self, p_Type, p_Value, p_Traceback):
            self.release()

    class _aWriter(object):
        def __init__(self, p_RWLock):
            self.A_RWLock = p_RWLock
            self.V_Locked = False

        def acquire(self, blocking=1, timeout=-1):
            self.V_Locked = self.A_RWLock.A_Resource.acquire(blocking, timeout)
            return self.V_Locked

        def release(self):
            if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
            self.V_Locked = False
            self.A_RWLock.A_Resource.release()

        def locked(self):
            return self.V_Locked

        def __enter__(self):
            self.acquire()

        def __exit__(self, p_Type, p_Value, p_Traceback):
            self.release()

    def genRlock(self):
        """
            Generate a reader lock
        """
        return RWLockRead._aReader(self)

    def genWlock(self):
        """
            Generate a writer lock
        """
        return RWLockRead._aWriter(self)


class RWLockWrite(object):
    """
        A Read/Write lock giving preference to Writer
    """

    def __init__(self):
        self.V_ReadCount = 0
        self.V_WriteCount = 0
        self.A_LockReadCount = threading.Lock()
        self.A_LockWriteCount = threading.Lock()
        self.A_LockReadEntry = threading.Lock()
        self.A_LockReadTry = threading.Lock()
        self.A_Resource = threading.Lock()

    class _aReader(object):
        def __init__(self, p_RWLock):
            self.A_RWLock = p_RWLock
            self.V_Locked = False

        def acquire(self, blocking=1, timeout=-1):
            p_TimeOut = None if (blocking and timeout < 0) else (timeout if blocking else 0)
            c_DeadLine = None if p_TimeOut is None else (time.time() + p_TimeOut)
            if not self.A_RWLock.A_LockReadEntry.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                               c_DeadLine - time.time())):
                return False
            if not self.A_RWLock.A_LockReadTry.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                             c_DeadLine - time.time())):
                self.A_RWLock.A_LockReadEntry.release()
                return False
            if not self.A_RWLock.A_LockReadCount.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                               c_DeadLine - time.time())):
                self.A_RWLock.A_LockReadTry.release()
                self.A_RWLock.A_LockReadEntry.release()
                return False
            self.A_RWLock.V_ReadCount += 1
            if (self.A_RWLock.V_ReadCount == 1):
                if not self.A_RWLock.A_Resource.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                              c_DeadLine - time.time())):
                    self.A_RWLock.A_LockReadTry.release()
                    self.A_RWLock.A_LockReadEntry.release()
                    self.A_RWLock.V_ReadCount -= 1
                    self.A_RWLock.A_LockReadCount.release()
                    return False
            self.A_RWLock.A_LockReadCount.release()
            self.A_RWLock.A_LockReadTry.release()
            self.A_RWLock.A_LockReadEntry.release()
            self.V_Locked = True
            return True

        def release(self):
            if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
            self.V_Locked = False
            self.A_RWLock.A_LockReadCount.acquire()
            self.A_RWLock.V_ReadCount -= 1
            if (self.A_RWLock.V_ReadCount == 0):
                self.A_RWLock.A_Resource.release()
            self.A_RWLock.A_LockReadCount.release()

        def locked(self):
            return self.V_Locked

        def __enter__(self):
            self.acquire()

        def __exit__(self, p_Type, p_Value, p_Traceback):
            self.release()

    class _aWriter(object):
        def __init__(self, p_RWLock):
            self.A_RWLock = p_RWLock
            self.V_Locked = False

        def acquire(self, blocking=1, timeout=-1):
            p_TimeOut = None if (blocking and timeout < 0) else (timeout if blocking else 0)
            c_DeadLine = None if p_TimeOut is None else (time.time() + p_TimeOut)
            if not self.A_RWLock.A_LockWriteCount.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                                c_DeadLine - time.time())):
                return False
            self.A_RWLock.V_WriteCount += 1
            if (self.A_RWLock.V_WriteCount == 1):
                if not self.A_RWLock.A_LockReadTry.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                                 c_DeadLine - time.time())):
                    self.A_RWLock.V_WriteCount -= 1
                    self.A_RWLock.A_LockWriteCount.release()
                    return False
            self.A_RWLock.A_LockWriteCount.release()
            if not self.A_RWLock.A_Resource.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                          c_DeadLine - time.time())):
                self.A_RWLock.A_LockWriteCount.acquire()
                self.A_RWLock.V_WriteCount -= 1
                if self.A_RWLock.V_WriteCount == 0:
                    self.A_RWLock.A_LockReadTry.release()
                self.A_RWLock.A_LockWriteCount.release()
                return False
            self.V_Locked = True
            return True

        def release(self):
            if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
            self.V_Locked = False
            self.A_RWLock.A_Resource.release()
            self.A_RWLock.A_LockWriteCount.acquire()
            self.A_RWLock.V_WriteCount -= 1
            if (self.A_RWLock.V_WriteCount == 0):
                self.A_RWLock.A_LockReadTry.release()
            self.A_RWLock.A_LockWriteCount.release()

        def locked(self):
            return self.V_Locked

        def __enter__(self):
            self.acquire()

        def __exit__(self, p_Type, p_Value, p_Traceback):
            self.release()

    def genRlock(self):
        """
            Generate a reader lock
        """
        return RWLockWrite._aReader(self)

    def genWlock(self):
        """
            Generate a writer lock
        """
        return RWLockWrite._aWriter(self)


class RWLockFair(object):
    """
        A Read/Write lock giving fairness to both Reader and Writer
    """

    def __init__(self):
        self.V_ReadCount = 0
        self.A_LockReadCount = threading.Lock()
        self.A_LockRead = threading.Lock()
        self.A_LockWrite = threading.Lock()

    class _aReader(object):
        def __init__(self, p_RWLock):
            self.A_RWLock = p_RWLock
            self.V_Locked = False

        def acquire(self, blocking=1, timeout=-1):
            p_TimeOut = None if (blocking and timeout < 0) else (timeout if blocking else 0)
            c_DeadLine = None if p_TimeOut is None else (time.time() + p_TimeOut)
            if not self.A_RWLock.A_LockRead.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                          c_DeadLine - time.time())):
                return False
            if not self.A_RWLock.A_LockReadCount.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                               c_DeadLine - time.time())):
                self.A_RWLock.A_LockRead.release()
                return False
            self.A_RWLock.V_ReadCount += 1
            if self.A_RWLock.V_ReadCount == 1:
                if not self.A_RWLock.A_LockWrite.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                               c_DeadLine - time.time())):
                    self.A_RWLock.V_ReadCount -= 1
                    self.A_RWLock.A_LockReadCount.release()
                    self.A_RWLock.A_LockRead.release()
                    return False
            self.A_RWLock.A_LockReadCount.release()
            self.A_RWLock.A_LockRead.release()
            self.V_Locked = True
            return True

        def release(self):
            if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
            self.V_Locked = False
            self.A_RWLock.A_LockReadCount.acquire()
            self.A_RWLock.V_ReadCount -= 1
            if self.A_RWLock.V_ReadCount == 0:
                self.A_RWLock.A_LockWrite.release()
            self.A_RWLock.A_LockReadCount.release()

        def locked(self):
            return self.V_Locked

        def __enter__(self):
            self.acquire()

        def __exit__(self, p_Type, p_Value, p_Traceback):
            self.release()

    class _aWriter(object):
        def __init__(self, p_RWLock):
            self.A_RWLock = p_RWLock
            self.V_Locked = False

        def acquire(self, blocking=1, timeout=-1):
            p_TimeOut = None if (blocking and timeout < 0) else (timeout if blocking else 0)
            c_DeadLine = None if p_TimeOut is None else (time.time() + p_TimeOut)
            if not self.A_RWLock.A_LockRead.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                          c_DeadLine - time.time())):
                return False
            if not self.A_RWLock.A_LockWrite.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0,
                                                                                                           c_DeadLine - time.time())):
                self.A_RWLock.A_LockRead.release()
                return False
            self.V_Locked = True
            return True

        def release(self):
            if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
            self.V_Locked = False
            self.A_RWLock.A_LockWrite.release()
            self.A_RWLock.A_LockRead.release()

        def locked(self):
            return self.V_Locked

        def __enter__(self):
            self.acquire()

        def __exit__(self, p_Type, p_Value, p_Traceback):
            self.release()

    def genRlock(self):
        """
            Generate a reader lock
        """
        return RWLockFair._aReader(self)

    def genWlock(self):
        """
            Generate a writer lock
        """
        return RWLockFair._aWriter(self)



# IGMP timers (in seconds)
ROBUSTNESS_VARIABLE = 2
QUERY_INTERVAL = 125
QUERY_RESPONSE_INTERVAL = 10
MAX_RESPONSE_TIME_QUERY_RESPONSE_INTERVAL = QUERY_RESPONSE_INTERVAL * 10
GROUP_MEMBERSHIP_INTERVAL = ROBUSTNESS_VARIABLE * QUERY_INTERVAL + QUERY_RESPONSE_INTERVAL
OTHER_QUERIER_PRESENT_INTERVAL = ROBUSTNESS_VARIABLE * QUERY_INTERVAL + QUERY_RESPONSE_INTERVAL / 2
STARTUP_QUERY_INTERVAL = QUERY_INTERVAL / 4
STARTUP_QUERY_COUNT = ROBUSTNESS_VARIABLE
LAST_MEMBER_QUERY_INTERVAL = 1
MAX_RESPONSE_TIME_LAST_MEMBER_QUERY_INTERVAL = LAST_MEMBER_QUERY_INTERVAL * 10
LAST_MEMBER_QUERY_COUNT = ROBUSTNESS_VARIABLE
UNSOLICITED_REPORT_INTERVAL = 10
VERSION_1_ROUTER_PRESENT_TIMEOUT = 400

# IGMP msg type
MEMBERSHIP_QUERY = 0x11
VERSION_1_MEMBERSHIP_REPORT = 0x12
VERSION_2_MEMBERSHIP_REPORT = 0x16
LEAVE_GROUP = 0x17



import sys
import time
import hashlib
import logging
import netifaces
import logging.handlers
from prettytable import PrettyTable
from hpimdm.TestLogger import RootFilter

from hpimdm.tree import hpim_globals
from hpimdm import UnicastRouting

interfaces = {}  # interfaces with multicast routing protocol enabled
interfaces_v6 = {}  # hpim ipv6 interfaces
igmp_interfaces = {}  # igmp interfaces
mld_interfaces = {}  # mld interfaces
kernel = None
kernel_v6 = None
unicast_routing = None
logger = None


def add_hpim_interface(interface_name, ipv4=True, ipv6=False):
    """
    Add a new interface to be controlled by HPIM-DM
    """
    if interface_name == "*":
        for interface_name in netifaces.interfaces():
            add_hpim_interface(interface_name, ipv4, ipv6)
        return

    if ipv4 and kernel is not None:
        kernel.create_hpim_interface(interface_name=interface_name)
    if ipv6 and kernel_v6 is not None:
        kernel_v6.create_hpim_interface(interface_name=interface_name)


def add_membership_interface(interface_name, ipv4=True, ipv6=False):
    """
    Add a new interface to be controlled by IGMP/MLD
    """
    if interface_name == "*":
        for interface_name in netifaces.interfaces():
            add_membership_interface(interface_name, ipv4, ipv6)
        return

    if ipv4 and kernel is not None:
        kernel.create_membership_interface(interface_name=interface_name)
    if ipv6 and kernel_v6 is not None:
        kernel_v6.create_membership_interface(interface_name=interface_name)


def remove_interface(interface_name, hpim=False, membership=False, ipv4=True, ipv6=False):
    """
    Remove HPIM-DM/IGMP interface
    """
    if interface_name == "*":
        for interface_name in netifaces.interfaces():
            remove_interface(interface_name, hpim, membership, ipv4, ipv6)
        return

    if ipv4 and kernel is not None:
        kernel.remove_interface(interface_name, hpim=hpim, membership=membership)
    if ipv6 and kernel_v6 is not None:
        kernel_v6.remove_interface(interface_name, hpim=hpim, membership=membership)


def list_neighbors(ipv4=False, ipv6=False):
    """
    List all neighbors in a human readable format
    """
    if ipv4:
        interfaces_list = interfaces.values()
    elif ipv6:
        interfaces_list = interfaces_v6.values()
    else:
        return "Unknown IP family"

    t = PrettyTable(['Interface', 'Neighbor IP', 'State', 'Hello Hold Time', 'BootTime', 'NeighborSnapshotSN', 'Uptime'])
    check_time = time.time()
    for interface in interfaces_list:
        for neighbor in interface.get_neighbors():
            uptime = check_time - neighbor.time_of_last_update
            uptime = 0 if (uptime < 0) else uptime

            t.add_row(
                [interface.interface_name, neighbor.ip, neighbor.neighbor_state.__name__, neighbor.hello_hold_time,
                 neighbor.time_of_boot, neighbor.neighbor_snapshot_sn, time.strftime("%H:%M:%S", time.gmtime(uptime))])
    print(t)
    return str(t)


def list_enabled_interfaces(ipv4=False, ipv6=False):
    """
    List all interfaces of the machine (enabled and not enabled for HPIM-DM and IGMP)
    """
    if ipv4:
        t = PrettyTable(['Interface', 'IP', 'HPIM/IGMP Enabled', 'HPIM security', 'IGMP State'])
        family = netifaces.AF_INET
        hpim_interfaces = interfaces
        membership_interfaces = igmp_interfaces
    elif ipv6:
        t = PrettyTable(['Interface', 'IP', 'HPIM/MLD Enabled', 'HPIM security', 'MLD State'])
        family = netifaces.AF_INET6
        hpim_interfaces = interfaces_v6
        membership_interfaces = mld_interfaces
    else:
        return "Unknown IP family"

    for interface in netifaces.interfaces():
        try:
            # TODO: fix same interface with multiple ips
            ip = netifaces.ifaddresses(interface)[family][0]['addr']
            hpim_enabled = interface in hpim_interfaces
            membership_enabled = interface in membership_interfaces
            enabled = str(hpim_enabled) + "/" + str(membership_enabled)
            hpim_protection = "-"
            if hpim_enabled and hpim_interfaces[interface].is_security_enabled():
                hpim_protection = str(hpim_interfaces[interface].security_id) + ": " + \
                                  hpim_interfaces[interface].hash_function.__name__
            membership_state = "-"
            if membership_enabled:
                membership_state = membership_interfaces[interface].interface_state.print_state()
            t.add_row([interface, ip, enabled, hpim_protection, membership_state])
        except Exception:
            continue
    print(t)
    return str(t)


def list_state(ipv4=True, ipv6=False):
    """
    List IGMP and HPIM-DM state
    For IGMP/MLD list the state of each group, regarding each interface
    For HPIM-DM list all trees and state of each interface
    """
    state_text = ""
    if ipv4:
        state_text = "IGMP State:\n{}\n\n\n\nMulticast Routing State:\n{}"
    elif ipv6:
        state_text = "MLD State:\n{}\n\n\n\nMulticast Routing State:\n{}"
    else:
        return state_text

    return state_text.format(list_membership_state(ipv4, ipv6), list_routing_state(ipv4, ipv6))


def list_membership_state(ipv4=True, ipv6=False):
    """
    List IGMP/MLD state (state of each group regarding each interface)
    """
    t = PrettyTable(['Interface', 'RouterState', 'Group Adress', 'GroupState'])
    if ipv4 and igmp_interfaces is not None:
        membership_interfaces = igmp_interfaces
    elif ipv6 and mld_interfaces is not None:
        membership_interfaces = mld_interfaces
    else:
        membership_interfaces = {}

    for (interface_name, interface_obj) in list(membership_interfaces.items()):
        interface_state = interface_obj.interface_state
        state_txt = interface_state.print_state()
        print(interface_state.group_state.items())

        for (group_addr, group_state) in list(interface_state.group_state.items()):
            print(group_addr)
            group_state_txt = group_state.print_state()
            t.add_row([interface_name, state_txt, group_addr, group_state_txt])
    return str(t)


def list_routing_state(ipv4=False, ipv6=False):
    """
    List HPIM-DM state (all state machines of each tree, regarding each interface)
    """
    if ipv4 and kernel is not None:
        routes = kernel.routing.values()
        vif_indexes = kernel.vif_index_to_name_dic.keys()
        dict_index_to_name = kernel.vif_index_to_name_dic
    elif ipv6 and kernel_v6 is not None:
        routes = kernel_v6.routing.values()
        vif_indexes = kernel_v6.vif_index_to_name_dic.keys()
        dict_index_to_name = kernel_v6.vif_index_to_name_dic
    else:
        routes = []
        vif_indexes = []
        dict_index_to_name = {}

    routing_entries = []
    for a in list(routes):
        for b in list(a.values()):
            routing_entries.append(b)

    t = PrettyTable(['SourceIP', 'GroupIP', 'TreeState', 'Interface', 'InterestState', 'AssertState', 'LocalMembership', "Is Forwarding?"])
    for entry in routing_entries:
        ip = entry.source_ip
        group = entry.group_ip
        upstream_if_index = entry.inbound_interface_index
        tree_state = str(entry._tree_state)

        for index in vif_indexes:
            interface_state = entry.interface_state[index]
            interface_name = dict_index_to_name[index]
            local_membership = str(interface_state._local_membership_state)
            try:
                if index != upstream_if_index:
                    assert_state = str(interface_state._assert_state)
                    prune_state = str(interface_state._downstream_node_interest_state)
                    is_forwarding = interface_state.is_forwarding()
                else:
                    assert_state = "--"
                    prune_state = "--"
                    is_forwarding = "root"
            except:
                prune_state = "-"
                assert_state = "-"
                is_forwarding = "-"

            t.add_row([ip, group, tree_state, interface_name, prune_state, assert_state, local_membership, is_forwarding])
    return str(t)


def list_sequence_numbers(ipv4=False, ipv6=False):
    """
    List all stored sequence numbers
    """
    if ipv4 and interfaces is not None:
        hpim_interfaces = interfaces
    elif ipv6 and interfaces_v6 is not None:
        hpim_interfaces = interfaces_v6
    else:
        raise Exception("Unknown IP family")

    t = PrettyTable(['Interface', 'BootTime', 'InterfaceSN'])
    for (interface_name, interface_obj) in hpim_interfaces.copy().items():
        interface_boot_time = interface_obj.time_of_boot
        interface_sn = interface_obj.sequencer
        t.add_row([interface_name, interface_boot_time, interface_sn])
    table_txt = str(t) + "\n\n\n"

    t = PrettyTable(['Neighbor', 'BootTime', 'NeighborSnapshotSN', 'NeighborCheckpointSN'])
    for interface in hpim_interfaces.values():
        for neighbor in interface.get_neighbors():
            t.add_row([neighbor.ip, neighbor.time_of_boot, neighbor.neighbor_snapshot_sn, neighbor.checkpoint_sn])
    table_txt += str(t) + "\n\n"

    t = PrettyTable(['Neighbor', 'Tree', 'LastSN'])
    for interface in hpim_interfaces.values():
        for neighbor in interface.get_neighbors():
            neighbor_last_sn = neighbor.last_sequence_number.copy()
            for (tree, last_sn) in neighbor_last_sn.items():
                t.add_row([neighbor.ip, tree, last_sn])
    table_txt += str(t)
    return str(table_txt)


def list_neighbors_state(ipv4=False, ipv6=False):
    """
    List all stored state regarding each tree of each neighbor (Upstream and Interest state)
    """
    if ipv4 and interfaces is not None:
        hpim_interfaces = list(interfaces.values())
    elif ipv6 and interfaces_v6 is not None:
        hpim_interfaces = list(interfaces_v6.values())
    else:
        raise Exception("Unknown IP family")

    t = PrettyTable(['Interface', 'Neighbor', 'Tree', 'RPC Upstream State'])
    for interface in hpim_interfaces:
        for neighbor in interface.get_neighbors():
            for (tree_id, tree_state) in neighbor.tree_metric_state.copy().items():
                t.add_row([interface.interface_name, neighbor.ip, tree_id, tree_state])
    table_txt = "Upstream state:\n" + str(t)

    t = PrettyTable(['Interface', 'Neighbor', 'Tree', 'Interest State'])
    for interface in hpim_interfaces:
        for neighbor in interface.get_neighbors():
            for (tree_id, tree_state) in neighbor.tree_interest_state.copy().items():
                t.add_row([interface.interface_name, neighbor.ip, tree_id, tree_state])
    table_txt += "\n\n\nInterest state:\n" + str(t)
    return str(table_txt)


def list_hash_algorithms():
    """
    List compatible algorithms to be used on HMAC protection
    """
    t = PrettyTable(['HMAC Hash Algorithms'])
    for alg in hashlib.algorithms_guaranteed:
        t.add_row([alg])
    return str(t)


def add_security_key(interface_name, security_identifier, hash_function, key, ipv4=False, ipv6=False):
    """
    Enable HMAC protection in interface_name with security_identifier, HMAC based on hash_function and
    HMAC protection key
    """
    if ipv4 and kernel is not None:
        kernel.add_interface_security(interface_name, security_identifier, hash_function, key)
    elif ipv6 and kernel_v6 is not None:
        kernel_v6.add_interface_security(interface_name, security_identifier, hash_function, key)


def remove_security_key(interface_name, security_identifier, ipv4=False, ipv6=False):
    """
    Disable HMAC protection identified by security_identifier on interface interface_name
    """
    if ipv4 and kernel is not None:
        kernel.remove_interface_security(interface_name, security_identifier)
    elif ipv6 and kernel_v6 is not None:
        kernel_v6.remove_interface_security(interface_name, security_identifier)


def change_initial_flood_setting():
    """
    Change Initial Flood Setting, used to control the implicit interest of all neighbors
    (in order to flood or ignore initial packets)
    """
    hpim_globals.INITIAL_FLOOD_ENABLED ^= True
    if kernel is not None:
        kernel.recheck_all_trees_in_all_interfaces()
    if kernel_v6 is not None:
        kernel_v6.recheck_all_trees_in_all_interfaces()

    return "Flood is enabled?: " + str(hpim_globals.INITIAL_FLOOD_ENABLED)


def hold_forwarding_state():
    """
    Change Hold Forwarding State Setting, used to hold the forwarding state of an interface that was AW and that became
    AL
    """
    hpim_globals.AL_HOLD_FORWARDING_STATE_ENABLED ^= True
    return "Hold Forwarding State is enabled?: " + str(hpim_globals.AL_HOLD_FORWARDING_STATE_ENABLED)


def stop():
    """
    Stop process
    """
    remove_interface("*", hpim=True, membership=True, ipv4=True, ipv6=True)
    if kernel is not None:
        kernel.exit()
    if kernel_v6 is not None:
        kernel_v6.exit()
    unicast_routing.stop()


def test(router_name, server_logger_ip):
    """
    Test setting.... Used to send logs to a remote server
    """
    global logger
    socketHandler = logging.handlers.SocketHandler(server_logger_ip,
                                                   logging.handlers.DEFAULT_TCP_LOGGING_PORT)
    # don't bother with a formatter, since a socket handler sends the event as
    # an unformatted pickle
    socketHandler.addFilter(RootFilter(router_name))
    logger.addHandler(socketHandler)


def enable_ipv6_kernel():
    """
    Function to explicitly enable IPv6 Multicast Routing stack.
    This may not be enabled by default due to some old linux kernels that may not have IPv6 stack or do not have
    IPv6 multicast routing support
    """
    global kernel_v6
    from hpimdm.Kernel import Kernel6
    kernel_v6 = Kernel6()

    global interfaces_v6
    global mld_interfaces
    interfaces_v6 = kernel_v6.hpim_interface
    mld_interfaces = kernel_v6.membership_interface


def main():
    """
    Start process
    """
    # logging
    global logger
    logger = logging.getLogger('hpim')
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))

    global kernel
    from hpimdm.Kernel import Kernel4
    kernel = Kernel4()

    global unicast_routing
    unicast_routing = UnicastRouting.UnicastRouting()

    global interfaces
    global igmp_interfaces
    interfaces = kernel.hpim_interface
    igmp_interfaces = kernel.membership_interface

    try:
        enable_ipv6_kernel()
    except:
        pass



from hpimdm.utils import TYPE_CHECKING

if TYPE_CHECKING:
    from .tree_if_non_root import TreeInterfaceNonRoot


class SFMRNonRootState:
    @staticmethod
    def interfaces_roles_dont_change_and_tree_transitions_to_active_state(interface: 'TreeInterfaceNonRoot'):
        """
        Tree transitions to Active state AND
        interface not directly connected to the source AND
        interface roles dont change
        """
        interface.logger.debug('interfaces_roles_dont_change_and_tree_transitions_to_active_state')
        interface.send_i_am_upstream()

    @staticmethod
    def interfaces_roles_change_and_tree_remains_or_transitions_to_active_state(interface: 'TreeInterfaceNonRoot'):
        """
        Tree remains or changes to Active state AND
        interface not directly connected to the source AND
        interface roles change (Root->Non-Root)
        """
        interface.logger.debug('interfaces_roles_change_and_tree_remains_or_transitions_to_active_state')
        interface.send_i_am_upstream()

    @staticmethod
    def tree_transitions_from_active_to_unsure_and_best_upstream_neighbor_is_null(interface: 'TreeInterfaceNonRoot'):
        """
        Tree transitions from Active to Unsure state AND
        BestUpstream neighbor is null
        """
        interface.logger.debug('tree_transitions_from_active_to_unsure_and_best_upstream_neighbor_is_null')
        interface.send_i_am_no_longer_upstream()

    @staticmethod
    def tree_transitions_from_active_to_inactive(interface: 'TreeInterfaceNonRoot'):
        """
        Tree transitions from Active to Inactive state
        """
        interface.logger.debug('tree_transitions_from_active_to_inactive')
        interface.send_i_am_no_longer_upstream()

    @staticmethod
    def interface_roles_dont_change_and_tree_transitions_from_active_to_unsure_and_best_upstream_neighbor_is_not_null(interface: 'TreeInterfaceNonRoot'):
        """
        Tree transitions from Active to Unsure state AND
        BestUpstream neighbor is not null
        """
        interface.logger.debug('interface_roles_dont_change_and_tree_transitions_from_active_to_unsure_and_best_upstream_neighbor_is_not_null')
        interface.send_i_am_no_longer_upstream()
        interface.send_no_interest()

    @staticmethod
    def interface_roles_change_and_tree_transitions_from_active_to_unsure_and_best_upstream_neighbor_is_not_null(interface: 'TreeInterfaceNonRoot'):
        """
        Tree transitions from Active to Unsure state AND
        BestUpstream neighbor is not null
        """
        interface.logger.debug('interface_roles_change_and_tree_transitions_from_active_to_unsure_and_best_upstream_neighbor_is_not_null')
        interface.send_no_interest()

    @staticmethod
    def tree_remains_unsure_and_best_upstream_router_reelected(interface: 'TreeInterfaceNonRoot'):
        """
        Tree remains in Unsure state AND
        BestUpstream neighbor changes

        OR

        Tree remains in Unsure state AND
        BestUpstream neighbor transmitted IamUpstream message that doesnt cause a change of the BestUpstream neighbor
        """
        interface.logger.debug('tree_remains_unsure_and_best_upstream_router_reelected')
        interface.send_no_interest()

    @staticmethod
    def tree_transitions_from_inactive_to_unsure_and_best_upstream_is_not_null(interface: 'TreeInterfaceNonRoot'):
        """
        Tree transitions from Inactive to Unsure state AND
        BestUpstream neighbor is not null
        """
        interface.logger.debug('tree_transitions_from_inactive_to_unsure_and_best_upstream_is_not_null')
        interface.send_no_interest()

    @staticmethod
    def tree_remains_active_and_my_rpc_changes(interface: 'TreeInterfaceNonRoot'):
        """
        Tree is Active AND
        interface not directly connected to the source AND
        MyRPC changes
        """
        interface.logger.debug('tree_remains_active_and_my_rpc_changes')
        interface.send_i_am_upstream()
